# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Union, Tuple

from ... import oscar as mo
from ...lib.aio import AioFileObject
from ...oscar.backends.allocate_strategy import IdleLabel, NoIdleSlot
from ...resource import cuda_card_stats
from ...storage import StorageLevel, get_storage_backend
from ...storage.base import ObjectInfo, StorageBackend
from ...storage.core import StorageFileObject
from ...utils import dataslots
from .errors import DataNotExist, StorageFull

logger = logging.getLogger(__name__)


def build_data_info(storage_info: ObjectInfo, level, size, band_name=None):
    # todo handle multiple
    if band_name is None:
        band_name = (
            "numa-0" if storage_info.device is None else f"gpu-{storage_info.device}"
        )
    if storage_info.size is None:
        store_size = size
    else:
        store_size = storage_info.size
    return DataInfo(storage_info.object_id, level, size, store_size, band_name)


class WrappedStorageFileObject(AioFileObject):
    """
    Wrap to hold ref after write close
    """

    def __init__(
        self,
        file: StorageFileObject,
        level: StorageLevel,
        size: int,
        session_id: str,
        data_key: str,
        data_manager: mo.ActorRefType["DataManagerActor"],
        storage_handler: StorageBackend,
    ):
        self._object_id = file.object_id
        super().__init__(file)
        self._size = size
        self._level = level
        self._session_id = session_id
        self._data_key = data_key
        self._data_manager = data_manager
        self._storage_handler = storage_handler

    def __getattr__(self, item):
        return getattr(self._file, item)

    async def clean_up(self):
        self._file.close()

    async def close(self):
        self._file.close()
        if self._object_id is None:
            # for some backends like vineyard,
            # object id is generated after write close
            self._object_id = self._file.object_id
        if "w" in self._file.mode:
            object_info = await self._storage_handler.object_info(self._object_id)
            data_info = build_data_info(object_info, self._level, self._size)
            await self._data_manager.put_data_info(
                self._session_id, self._data_key, data_info, object_info
            )


class StorageQuotaActor(mo.Actor):
    def __init__(
        self,
        data_manager: mo.ActorRefType["DataManagerActor"],
        level: StorageLevel,
        total_size: Optional[Union[int, float]],
    ):
        self._data_manager = data_manager
        self._total_size = total_size if total_size is None else total_size * 0.95
        self._used_size = 0
        self._level = level

    @classmethod
    def gen_uid(cls, band_name: str, level: StorageLevel):
        return f"storage_quota_{band_name}_{level}"

    def update_quota(self, size: int):
        self._used_size += size
        logger.debug(
            "Update %s bytes of %s, used size now is %s",
            size,
            self._level,
            self._used_size,
        )

    def request_quota(self, size: int) -> bool:
        if self._total_size is not None and size > self._total_size:  # pragma: no cover
            raise StorageFull(
                f"Request size {size} is larger than total size {self._total_size}"
            )
        if self._total_size is not None and self._used_size + size > self._total_size:
            logger.debug(
                "Request %s bytes of %s, used size now is %s,"
                "space is not enough for the request",
                size,
                self._level,
                self._used_size,
            )
            return False
        else:
            self._used_size += size
            logger.debug(
                "Request %s bytes of %s, used size now is %s, total size is %s",
                size,
                self._level,
                self._used_size,
                self._total_size,
            )
            return True

    def release_quota(self, size: int):
        self._used_size -= size
        logger.debug(
            "Release %s bytes of %s, used size now is %s, total size is %s",
            size,
            self._level,
            self._used_size,
            self._total_size,
        )

    def get_quota(self) -> Tuple[float, float]:
        return self._total_size, self._used_size


@dataslots
@dataclass
class DataInfo:
    object_id: object
    level: StorageLevel
    memory_size: int
    store_size: int
    band: str = None


@dataslots
@dataclass
class InternalDataInfo:
    data_info: DataInfo
    object_info: ObjectInfo


class DataManagerActor(mo.Actor):
    _data_key_to_info: Dict[tuple, List[InternalDataInfo]]

    def __init__(self, bands: List):
        from .spill import FIFOStrategy

        # mapping key is (session_id, data_key)
        # mapping value is list of InternalDataInfo
        self._bands = bands
        self._data_key_to_info = defaultdict(list)
        self._data_info_list = dict()
        self._spill_strategy = dict()
        # data key may be a tuple in shuffle cases,
        # we record the mapping from main key to sub keys
        self._main_key_to_sub_keys = defaultdict(set)
        for level in StorageLevel.__members__.values():
            for band_name in bands:
                self._data_info_list[level, band_name] = dict()
                self._spill_strategy[level, band_name] = FIFOStrategy(level)

    def _get_data_infos(
        self, session_id: str, data_key: str, band_name: str, error: str
    ) -> Union[List[DataInfo], None]:
        if (session_id, data_key) in self._data_key_to_info:
            available_infos = []
            for info in self._data_key_to_info[session_id, data_key]:
                info_band = info.data_info.band
                if info_band.startswith("gpu-"):  # pragma: no cover
                    # not available for different GPU bands
                    if info_band == band_name:
                        available_infos.append(info.data_info)
                else:
                    available_infos.append(info.data_info)
            return available_infos
        else:
            if error == "raise":
                raise DataNotExist(f"Data key {session_id, data_key} not exists.")
            else:
                return

    @mo.extensible
    def get_data_infos(
        self, session_id: str, data_key: str, band_name: str, error: str = "raise"
    ) -> Union[List[DataInfo], Dict]:
        if (session_id, data_key) in self._main_key_to_sub_keys:
            key_to_infos = dict()
            # if data_key is a main key, we return a dict contains all sub keys' info,
            # it occurs mostly when deleting all shuffle data using main key.
            for sub_key in self._main_key_to_sub_keys[(session_id, data_key)]:
                infos = self._get_data_infos(session_id, sub_key, band_name, error)
                if infos:
                    key_to_infos[sub_key] = infos
            return key_to_infos
        else:
            return self._get_data_infos(session_id, data_key, band_name, error)

    def _get_data_info(
        self, session_id: str, data_key: str, band_name: str, error: str = "raise"
    ) -> Union[DataInfo, None]:
        # if the data is stored in multiply levels,
        # return the lowest level info
        infos = self._get_data_infos(session_id, data_key, band_name, error)
        if not infos:
            return
        infos = sorted(infos, key=lambda x: x.level)
        return infos[0]

    @mo.extensible
    def get_data_info(
        self,
        session_id: str,
        data_key: str,
        band_name: str = None,
        error: str = "raise",
    ) -> Union[DataInfo, None]:
        return self._get_data_info(session_id, data_key, band_name, error)

    def _put_data_info(
        self,
        session_id: str,
        data_key: str,
        data_info: DataInfo,
        object_info: ObjectInfo = None,
    ):
        info = InternalDataInfo(data_info, object_info)
        self._data_key_to_info[(session_id, data_key)].append(info)
        self._data_info_list[data_info.level, data_info.band][
            (session_id, data_key)
        ] = object_info
        self._spill_strategy[data_info.level, data_info.band].record_put_info(
            (session_id, data_key), data_info.store_size
        )
        if isinstance(data_key, tuple):
            self._main_key_to_sub_keys[(session_id, data_key[0])].update([data_key])

    @mo.extensible
    def put_data_info(
        self,
        session_id: str,
        data_key: str,
        data_info: DataInfo,
        object_info: ObjectInfo = None,
    ):
        self._put_data_info(session_id, data_key, data_info, object_info=object_info)

    def _delete_data_info(
        self, session_id: str, data_key: str, level: StorageLevel, band_name: str
    ):
        if (session_id, data_key) in self._data_key_to_info:
            self._data_info_list[level, band_name].pop((session_id, data_key))
            self._spill_strategy[level, band_name].record_delete_info(
                (session_id, data_key)
            )
            infos = self._data_key_to_info[(session_id, data_key)]
            rest = [info for info in infos if info.data_info.level != level]
            if isinstance(data_key, tuple):
                if (
                    len(self._main_key_to_sub_keys[(session_id, data_key[0])]) == 0
                ):  # pragma: no cover
                    del self._main_key_to_sub_keys[(session_id, data_key[0])]
            if len(rest) == 0:
                del self._data_key_to_info[(session_id, data_key)]
            else:  # pragma: no cover
                self._data_key_to_info[(session_id, data_key)] = rest

    @mo.extensible
    def delete_data_info(
        self, session_id: str, data_key: str, level: StorageLevel, band_name: str
    ):
        self._delete_data_info(session_id, data_key, level, band_name)

    def list(self, level: StorageLevel, ban_name: str):
        return list(self._data_info_list[level, ban_name].keys())

    @mo.extensible
    def pin(self, session_id, data_key, band_name, error="raise"):
        info = self.get_data_info(session_id, data_key, band_name, error=error)
        if info is not None:
            self._spill_strategy[info.level, info.band].pin_data((session_id, data_key))

    @mo.extensible
    def unpin(
        self,
        session_id: str,
        data_keys: List[str],
        band_name: str,
        error: str = "raise",
    ):
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")
        levels = set()
        for data_key in data_keys:
            info = self.get_data_info(session_id, data_key, band_name, error)
            if info:
                level = info.level
                self._spill_strategy[level, info.band].unpin_data(
                    (session_id, data_key)
                )
                levels.add(level)
        return list(levels)

    def get_spillable_size(self, level: StorageLevel, band_name: str):
        return self._spill_strategy[level, band_name].get_spillable_size()

    async def get_spill_keys(self, level: StorageLevel, band_name: str, size: int):
        return self._spill_strategy[level, band_name].get_spill_keys(size)


class StorageManagerActor(mo.StatelessActor):
    """
    Storage manager actor, created only on main process, mainly to setup storage backends
    and create all the necessary actors for storage service.
    """

    _data_manager: mo.ActorRefType[DataManagerActor]

    def __init__(
        self, storage_configs: Dict, transfer_block_size: int = None, **kwargs
    ):
        from .handler import StorageHandlerActor

        self._handler_cls = kwargs.pop("storage_handler_cls", StorageHandlerActor)
        self._storage_configs = storage_configs
        self._all_bands = None
        self._cluster_api = None
        self._upload_task = None

        # params to init and teardown
        self._init_params = defaultdict(dict)
        self._teardown_params = defaultdict(dict)
        self._supervisor_address = None

        # transfer config
        self._transfer_block_size = transfer_block_size
        self._quotas = None
        self._spill_managers = None

    async def __post_create__(self):
        from ..cluster.api import ClusterAPI
        from .handler import StorageHandlerActor

        try:
            self._cluster_api = cluster_api = await ClusterAPI.create(self.address)
            band_to_resource = await cluster_api.get_bands()
            self._all_bands = [band[1] for band in band_to_resource]
        except mo.ActorNotExist:
            # in some test cases, cluster service is not available
            self._all_bands = ["numa-0"]

        # stores the mapping from data key to storage info
        self._data_manager = await mo.create_actor(
            DataManagerActor,
            self._all_bands,
            uid=DataManagerActor.default_uid(),
            address=self.address,
        )

        # setup storage backend
        await self._setup_storage_backends()

        # create in main process
        default_band_name = "numa-0"
        await mo.create_actor(
            self._handler_cls,
            self._init_params[default_band_name],
            self._data_manager,
            self._spill_managers[default_band_name],
            self._quotas[default_band_name],
            default_band_name,
            uid=StorageHandlerActor.gen_uid(default_band_name),
            address=self.address,
        )

        # create handler actors for every process
        await self._create_storage_handler_actors()
        # create actor for transfer
        await self._create_transfer_actors()
        await self.upload_disk_info()
        # create task for uploading storage usages
        self._upload_task = asyncio.create_task(self.upload_storage_info())

    async def __pre_destroy__(self):
        if self._upload_task:
            self._upload_task.cancel()
        for _, params in self._teardown_params.items():
            for backend, teardown_params in params.items():
                backend_cls = get_storage_backend(backend)
                await backend_cls.teardown(**teardown_params)

    async def _setup_storage_backends(self):
        from .spill import SpillManagerActor

        self._quotas = quotas = defaultdict(dict)
        self._spill_managers = spill_managers = defaultdict(dict)
        for backend, setup_params in self._storage_configs.items():
            if backend == "cuda":  # pragma: no cover
                cuda_infos = await asyncio.to_thread(cuda_card_stats)
                storage_bands = [s for s in self._all_bands if s.startswith("gpu-")]
                clients = []
                for gpu_band in storage_bands:
                    index = int(gpu_band[4:])
                    size = cuda_infos[index].fb_mem_info.available
                    params = dict(size=size, **setup_params)
                    clients.append(await self._setup_storage(gpu_band, backend, params))
            else:
                storage_bands = ["numa-0"]
                clients = [
                    await self._setup_storage(band_name, backend, setup_params)
                    for band_name in storage_bands
                ]

            for level in StorageLevel.__members__.values():
                for client, storage_band in zip(clients, storage_bands):
                    if client.level & level:
                        logger.debug(
                            "Create quota manager for %s, total size is %s",
                            level,
                            client.size,
                        )
                        quotas[storage_band][level] = await mo.create_actor(
                            StorageQuotaActor,
                            self._data_manager,
                            level,
                            client.size,
                            uid=StorageQuotaActor.gen_uid(storage_band, level),
                            address=self.address,
                        )
                        spill_managers[storage_band][level] = await mo.create_actor(
                            SpillManagerActor,
                            level,
                            uid=SpillManagerActor.gen_uid(storage_band, level),
                            address=self.address,
                        )

    async def _create_storage_handler_actors(self):
        from .handler import StorageHandlerActor
        from .transfer import SenderManagerActor, ReceiverManagerActor

        for band_name in self._init_params:
            strategy = IdleLabel(band_name, "StorageHandler")
            sender_strategy = IdleLabel(band_name, "sender")
            receiver_strategy = IdleLabel(band_name, "receiver")
            init_params = self._get_band_init_params(band_name)
            band_spill_managers = self._get_band_spill_managers(band_name)
            band_quotas = self._get_band_quota_refs(band_name)
            while True:
                try:
                    handler_ref = await mo.create_actor(
                        self._handler_cls,
                        init_params,
                        self._data_manager,
                        band_spill_managers,
                        band_quotas,
                        band_name,
                        uid=StorageHandlerActor.gen_uid(band_name),
                        address=self.address,
                        allocate_strategy=strategy,
                    )
                    # create transfer actor for GPU bands
                    if band_name.startswith("gpu-"):  # pragma: no cover
                        await mo.create_actor(
                            SenderManagerActor,
                            band_name,
                            data_manager_ref=self._data_manager,
                            storage_handler_ref=handler_ref,
                            uid=SenderManagerActor.gen_uid(band_name),
                            address=self.address,
                            allocate_strategy=sender_strategy,
                        )
                        await mo.create_actor(
                            ReceiverManagerActor,
                            band_quotas,
                            handler_ref,
                            address=self.address,
                            uid=ReceiverManagerActor.gen_uid(band_name),
                            allocate_strategy=receiver_strategy,
                        )
                except NoIdleSlot:
                    break

    async def _create_transfer_actors(self):
        from .handler import StorageHandlerActor
        from .transfer import SenderManagerActor, ReceiverManagerActor

        default_band_name = "numa-0"
        sender_strategy = IdleLabel("io", "sender")
        receiver_strategy = IdleLabel("io", "receiver")
        handler_strategy = IdleLabel("io", "handler")
        while True:
            try:
                handler_ref = await mo.create_actor(
                    self._handler_cls,
                    self._init_params[default_band_name],
                    self._data_manager,
                    self._spill_managers[default_band_name],
                    self._quotas[default_band_name],
                    default_band_name,
                    uid=StorageHandlerActor.gen_uid(default_band_name),
                    address=self.address,
                    allocate_strategy=handler_strategy,
                )
                await mo.create_actor(
                    SenderManagerActor,
                    data_manager_ref=self._data_manager,
                    storage_handler_ref=handler_ref,
                    uid=SenderManagerActor.gen_uid(default_band_name),
                    address=self.address,
                    allocate_strategy=sender_strategy,
                )

                await mo.create_actor(
                    ReceiverManagerActor,
                    self._quotas[default_band_name],
                    handler_ref,
                    address=self.address,
                    uid=ReceiverManagerActor.gen_uid(default_band_name),
                    allocate_strategy=receiver_strategy,
                )
            except NoIdleSlot:
                break

    def _get_band_init_params(self, band_name):
        init_params = self._init_params["numa-0"].copy()
        init_params.update(self._init_params[band_name])
        return init_params

    def _get_band_quota_refs(self, band_name):
        band_quotas = self._quotas[band_name].copy()
        band_quotas.update(self._quotas["numa-0"])
        return band_quotas

    def _get_band_spill_managers(self, band_name):
        band_spill_managers = self._spill_managers[band_name].copy()
        band_spill_managers.update(self._spill_managers["numa-0"])
        return band_spill_managers

    async def _setup_storage(
        self, band_name: str, storage_backend: str, storage_config: Dict
    ):
        backend = get_storage_backend(storage_backend)
        storage_config = storage_config or dict()
        init_params, teardown_params = await backend.setup(**storage_config)
        client = backend(**init_params)
        self._init_params[band_name][storage_backend] = init_params
        self._teardown_params[band_name][storage_backend] = teardown_params
        return client

    def get_client_params(self):
        return self._init_params

    async def upload_storage_info(self):
        from ..cluster import StorageInfo

        if self._cluster_api is not None:
            while True:
                upload_tasks = []
                for band, level_to_quota in self._quotas.items():
                    for level, quota_ref in level_to_quota.items():
                        total, used = await quota_ref.get_quota()
                        used = int(used)
                        if total is not None:
                            total = int(total)
                        storage_info = StorageInfo(
                            storage_level=level, total_size=total, used_size=used
                        )
                        upload_tasks.append(
                            self._cluster_api.set_band_storage_info.delay(
                                band, storage_info
                            )
                        )
                await self._cluster_api.set_band_storage_info.batch(*upload_tasks)
                await asyncio.sleep(0.5)

    async def upload_disk_info(self):
        from ..cluster import DiskInfo

        disk_infos = []
        if self._cluster_api is not None and "disk" in self._init_params["numa-0"]:
            params = self._init_params["numa-0"]["disk"]
            size = params["size"]
            for path in params["root_dirs"]:
                disk_infos.append(DiskInfo(path=path, limit_size=size))
            await self._cluster_api.set_node_disk_info(disk_infos)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import oscar as mo
from ...lib.aio import alru_cache
from ...oscar.backends.context import ProfilingContext
from ...oscar.profiling import MARS_ENABLE_PROFILING
from .core import Subtask


class SubtaskAPI:
    def __init__(self, address: str):
        self._address = address

    @classmethod
    async def create(cls, address: str) -> "SubtaskAPI":
        return SubtaskAPI(address)

    @alru_cache(cache_exceptions=False)
    async def _get_runner_ref(self, band_name: str, slot_id: int):
        from .worker.runner import SubtaskRunnerActor

        return await mo.actor_ref(
            SubtaskRunnerActor.gen_uid(band_name, slot_id), address=self._address
        )

    @alru_cache(cache_exceptions=False)
    async def _get_subtask_processor_ref(self, session_id: str, slot_address: str):
        from .worker.processor import SubtaskProcessorActor

        return await mo.actor_ref(
            SubtaskProcessorActor.gen_uid(session_id), address=slot_address
        )

    async def run_subtask_in_slot(self, band_name: str, slot_id: int, subtask: Subtask):
        """
        Run subtask in current worker

        Parameters
        ----------
        band_name
        subtask
        slot_id

        Returns
        -------

        """
        ref = await self._get_runner_ref(band_name, slot_id)
        extra_config = subtask.extra_config
        enable_profiling = MARS_ENABLE_PROFILING or (
            extra_config and extra_config.get("enable_profiling")
        )
        profiling_context = (
            ProfilingContext(task_id=subtask.task_id) if enable_profiling else None
        )
        return await ref.run_subtask.options(profiling_context=profiling_context).send(
            subtask
        )

    async def cancel_subtask_in_slot(self, band_name: str, slot_id: int):
        """
        Cancel subtask running in a worker slot and wait until it is cancelled

        Parameters
        ----------
        band_name : str
            name of a worker band, for instance, 'numa-0'
        slot_id : int
            index of a slot in a band
        """
        ref = await self._get_runner_ref(band_name, slot_id)
        await ref.cancel_subtask()

    async def set_running_operand_progress(
        self, session_id: str, op_key: str, slot_address: str, progress: float
    ):
        ref = await self._get_subtask_processor_ref(session_id, slot_address)
        await ref.set_running_op_progress(op_key, progress)


class MockSubtaskAPI(SubtaskAPI):
    @classmethod
    async def create(cls, address: str) -> "SubtaskAPI":
        from .worker.manager import SubtaskRunnerManagerActor

        await mo.create_actor(
            SubtaskRunnerManagerActor,
            address,
            None,
            uid=SubtaskRunnerManagerActor.default_uid(),
            address=address,
        )
        return await super().create(address)

    @classmethod
    async def cleanup(cls, address: str):
        from .worker.manager import SubtaskRunnerManagerActor

        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=SubtaskRunnerManagerActor.default_uid(), address=address
            )
        )



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...core.context import get_context
from ...serialization.serializables import Int32Field, StringField
from ...tensor.datasource.from_vineyard import resolve_vineyard_socket
from ...utils import calc_nsplits, has_unknown_shape, lazy_import
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


vineyard = lazy_import("vineyard")
vy_data_utils = lazy_import("vineyard.data.utils", rename="vy_data_utils")


class DataFrameFromVineyard(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_FROM_VINEYARD_CHUNK

    # generated columns for metadata
    generated_columns = ["id", "worker_address", "dtypes", "shape", "index", "columns"]

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    # ObjectID in vineyard
    object_id = StringField("object_id")

    # a dummy attr to make sure ops have different keys
    operator_index = Int32Field("operator_index")

    def __init__(self, vineyard_socket=None, object_id=None, **kw):
        super().__init__(
            vineyard_socket=vineyard_socket,
            object_id=object_id,
            _output_types=[OutputType.dataframe],
            **kw
        )

    def check_inputs(self, inputs):
        # no inputs
        if inputs and len(inputs) > 0:
            raise ValueError("DataFrame data source has no inputs")

    def _new_chunks(self, inputs, kws=None, **kw):
        shape = kw.get("shape", None)
        self.extra_params[
            "shape"
        ] = shape  # set shape to make the operand key different
        return super()._new_chunks(inputs, kws=kws, **kw)

    def _new_tileables(self, inputs, kws=None, **kw):
        shape = kw.get("shape", None)
        self.extra_params[
            "shape"
        ] = shape  # set shape to make the operand key different
        return super()._new_tileables(inputs, kws=kws, **kw)

    def __call__(self, shape, dtypes=None, index_value=None, columns_value=None):
        return self.new_dataframe(
            None,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def tile(cls, op):
        ctx = get_context()
        workers = ctx.get_worker_addresses()

        out_chunks = []
        dtypes = pd.Series(
            [np.dtype("O")] * len(cls.generated_columns), index=cls.generated_columns
        )
        for index, worker in enumerate(workers):
            chunk_op = op.copy().reset_key()
            chunk_op.expect_worker = worker
            chunk_op.operator_index = index
            out_chunk = chunk_op.new_chunk(
                [],
                dtypes=dtypes,
                shape=(1, len(cls.generated_columns)),
                index=(index, 0),
                index_value=parse_index(pd.RangeIndex(0, 1)),
                columns_value=parse_index(pd.Index(cls.generated_columns)),
            )
            out_chunks.append(out_chunk)

        new_op = op.copy().reset_key()
        return new_op.new_dataframes(
            op.inputs,
            shape=(np.nan, np.nan),
            dtypes=dtypes,
            chunks=out_chunks,
            nsplits=((np.nan,), (np.nan,)),
            # use the same value as `read_csv`
            index_value=parse_index(pd.RangeIndex(0, 1)),
            columns_value=parse_index(pd.Index(cls.generated_columns)),
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        meta = client.get_meta(vineyard.ObjectID(op.object_id))
        chunks, dtypes = [], None
        for idx in range(meta["partitions_-size"]):
            chunk_meta = meta["partitions_-%d" % idx]
            columns = pd.Index(vy_data_utils.from_json(chunk_meta["columns_"]))
            shape = (np.nan, len(columns))
            if not chunk_meta.islocal:
                continue
            if dtypes is None:
                dtypes = []
                for idx in range(len(columns)):
                    column_meta = chunk_meta["__values_-value-%d" % idx]
                    dtype = vy_data_utils.normalize_dtype(
                        column_meta["value_type_"],
                        column_meta.get("value_type_meta_", None),
                    )
                    dtypes.append(dtype)
                dtypes = pd.Series(dtypes, index=columns)
            chunk_index = (
                chunk_meta["partition_index_row_"],
                chunk_meta["partition_index_column_"],
            )
            # chunk: (chunk_id, worker_address, dtype, shape, index, columns)
            chunks.append(
                (
                    repr(chunk_meta.id),
                    ctx.worker_address,
                    dtypes,
                    shape,
                    chunk_index,
                    columns,
                )
            )

        ctx[op.outputs[0].key] = pd.DataFrame(chunks, columns=cls.generated_columns)


class DataFrameFromVineyardChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.TENSOR_FROM_VINEYARD_CHUNK

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    # ObjectID of chunk in vineyard
    object_id = StringField("object_id")

    def __init__(self, vineyard_socket=None, object_id=None, **kw):
        super().__init__(vineyard_socket=vineyard_socket, object_id=object_id, **kw)

    def __call__(self, meta):
        return self.new_dataframe([meta])

    @classmethod
    def tile(cls, op):
        if has_unknown_shape(*op.inputs):
            yield

        ctx = get_context()

        in_chunk_keys = [chunk.key for chunk in op.inputs[0].chunks]
        out_chunks = []
        chunk_map = dict()
        dtypes, columns = None, None
        for chunk, infos in zip(
            op.inputs[0].chunks, ctx.get_chunks_result(in_chunk_keys)
        ):
            for _, info in infos.iterrows():
                chunk_op = op.copy().reset_key()
                chunk_op.object_id = info["id"]
                chunk_op.expect_worker = info["worker_address"]
                dtypes = info["dtypes"]
                columns = info["columns"]
                shape = info["shape"]
                chunk_index = info["index"]
                chunk_map[chunk_index] = info["shape"]
                out_chunk = chunk_op.new_chunk(
                    [chunk],
                    shape=shape,
                    index=chunk_index,
                    dtypes=dtypes,
                    index_value=parse_index(pd.RangeIndex(0, -1)),
                    columns_value=parse_index(pd.Index(columns)),
                )
                out_chunks.append(out_chunk)

        nsplits = calc_nsplits(chunk_map)
        shape = [np.sum(nsplit) for nsplit in nsplits]
        new_op = op.copy().reset_key()
        return new_op.new_dataframes(
            op.inputs,
            shape=shape,
            dtypes=dtypes,
            chunks=out_chunks,
            nsplits=nsplits,
            index_value=parse_index(pd.RangeIndex(0, -1)),
            columns_value=parse_index(pd.Index(columns)),
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        client = vineyard.connect(socket)
        ctx[op.outputs[0].key] = client.get(vineyard.ObjectID(op.object_id))


def from_vineyard(df, vineyard_socket=None):
    if vineyard is not None and isinstance(df, vineyard.Object):  # pragma: no cover
        if "vineyard::GlobalDataFrame" not in df.typename:
            raise TypeError(
                "The input dataframe %r is not a vineyard' GlobalDataFrame" % df
            )
        object_id = df.id
    else:
        object_id = df
    if vineyard is not None and isinstance(object_id, vineyard.ObjectID):
        object_id = repr(object_id)
    metaop = DataFrameFromVineyard(
        vineyard_socket=vineyard_socket,
        object_id=object_id,
        dtype=np.dtype("byte"),
        gpu=None,
    )
    meta = metaop(
        shape=(np.nan,),
        dtypes=pd.Series([]),
        index_value=parse_index(pd.Index([])),
        columns_value=parse_index(pd.Index([])),
    )
    op = DataFrameFromVineyardChunk(
        vineyard_socket=vineyard_socket, object_id=object_id, gpu=None
    )
    return op(meta)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from typing import Any, Callable, Dict, List, Tuple, Type, Union

import numpy as np

try:
    from numpy.core._exceptions import UFuncTypeError
except ImportError:  # pragma: no cover
    UFuncTypeError = None

from ...metrics import Metrics
from ...typing import TileableType, ChunkType, OperandType
from ...utils import calc_data_size
from ..context import Context
from ..mode import is_eager_mode
from ..entity import (
    OutputType,
    ExecutableTuple,
    get_chunk_types,
    get_tileable_types,
    get_output_types,
    get_fetch_class,
)

_op_type_to_executor: Dict[Type[OperandType], Callable] = dict()
_op_type_to_size_estimator: Dict[Type[OperandType], Callable] = dict()


op_executed_number = Metrics.counter(
    "mars.operand.executed_number", "The number of executed operands.", ("op",)
)


class TileableOperandMixin:
    __slots__ = ()

    def check_inputs(self, inputs: List[TileableType]):
        if not inputs:
            return

        for inp in inputs:
            if inp is not None and inp._need_execution():
                raise ValueError(
                    f"{inp} has unknown dtypes, "
                    f"it must be executed first before {str(type(self))}"
                )

    @classmethod
    def _check_if_gpu(cls, inputs: List[TileableType]):
        if not inputs:
            return None
        true_num = 0
        for inp in inputs:
            op = getattr(inp, "op", None)
            if op is None or op.gpu is None:
                return None
            true_num += int(op.gpu)
        if true_num == len(inputs):
            return True
        elif true_num == 0:
            return False
        return None

    def _tokenize_output(self, output_idx: int, **kw):
        return f"{self._key}_{output_idx}"

    def _create_chunk(self, output_idx: int, index: Tuple[int], **kw) -> ChunkType:
        output_type = kw.pop("output_type", None) or self._get_output_type(output_idx)
        if not output_type:
            raise ValueError("output_type should be specified")

        if isinstance(output_type, (list, tuple)):
            output_type = output_type[output_idx]
        chunk_type, chunk_data_type = get_chunk_types(output_type)
        kw["_i"] = output_idx
        kw["op"] = self
        kw["index"] = index
        if output_type == OutputType.scalar:
            # tensor
            kw["order"] = "C_ORDER"

        # key of output chunks may only contain keys for its output ids
        if "_key" not in kw:
            kw["_key"] = self._tokenize_output(output_idx, **kw)

        data = chunk_data_type(**kw)
        return chunk_type(data)

    def _new_chunks(
        self, inputs: List[ChunkType], kws: List[Dict] = None, **kw
    ) -> List[ChunkType]:
        output_limit = kw.pop("output_limit", None)
        if output_limit is None:
            output_limit = self.output_limit
        if isinstance(output_limit, float) and kws:
            output_limit = len(kws)

        self.check_inputs(inputs)
        self._set_inputs(inputs)
        if self.gpu is None:
            self.gpu = self._check_if_gpu(self._inputs)
        if self._key is None:
            self._update_key()

        chunks = []
        for j in range(output_limit):
            create_chunk_kw = kw.copy()
            if kws:
                create_chunk_kw.update(kws[j])
            index = create_chunk_kw.pop("index", None)
            chunk = self._create_chunk(j, index, **create_chunk_kw)
            chunks.append(chunk)

        self.outputs = chunks
        if len(chunks) > 1:
            # for each output chunk, hold the reference to the other outputs
            # so that either no one or everyone are gc collected
            for j, t in enumerate(chunks):
                t.data._siblings = [c.data for c in chunks[:j] + chunks[j + 1 :]]
        return chunks

    def new_chunks(
        self, inputs: List[ChunkType], kws: List[Dict] = None, **kwargs
    ) -> List[ChunkType]:
        """
        Create chunks.

        A chunk is a node in a fine grained graph, all the chunk objects are created by
        calling this function, it happens mostly in tiles.
        The generated chunks will be set as this operand's outputs and each chunk will
        hold this operand as it's op.

        Parameters
        ----------
        inputs : list
            Input chunks.
        kws : dict
            Kwargs for each output.
        kwargs : dict
            common kwargs for all outputs

        Returns
        -------
        chunks : list
            Output chunks.

        .. note::
            It's a final method, do not override.
            Override the method `_new_chunks` if needed.
        """
        return self._new_chunks(inputs, kws=kws, **kwargs)

    def new_chunk(
        self, inputs: List[ChunkType], kws: List[Dict] = None, **kw
    ) -> ChunkType:
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new chunk with more than 1 outputs")

        return self.new_chunks(inputs, kws=kws, **kw)[0]

    @staticmethod
    def _fill_nan_shape(kw: dict):
        nsplits = kw.get("nsplits")
        shape = kw.get("shape")
        if nsplits is not None and shape is not None:
            nsplits = tuple(nsplits)
            shape = list(shape)
            for idx, (s, sp) in enumerate(zip(shape, nsplits)):
                if not np.isnan(s):
                    continue
                s = sum(sp)
                if not np.isnan(s):
                    shape[idx] = s
            kw["shape"] = tuple(shape)
            kw["nsplits"] = nsplits
        return kw

    def _create_tileable(self, output_idx: int, **kw) -> TileableType:
        output_type = kw.pop("output_type", self._get_output_type(output_idx))
        if output_type is None:
            raise ValueError("output_type should be specified")

        if isinstance(output_type, (list, tuple)):
            output_type = output_type[output_idx]

        tileable_type, tileable_data_type = get_tileable_types(output_type)
        kw["_i"] = output_idx
        kw["op"] = self
        if output_type == OutputType.scalar:
            # tensor
            kw["order"] = "C_ORDER"

        kw = self._fill_nan_shape(kw)

        # key of output chunks may only contain keys for its output ids
        if "_key" not in kw:
            kw["_key"] = self._tokenize_output(output_idx, **kw)

        data = tileable_data_type(**kw)
        return tileable_type(data)

    def _new_tileables(
        self, inputs: List[TileableType], kws: List[dict] = None, **kw
    ) -> List[TileableType]:
        output_limit = kw.pop("output_limit", None)
        if output_limit is None:
            output_limit = getattr(self, "output_limit")

        self._set_inputs(inputs)
        if self.gpu is None:
            self.gpu = self._check_if_gpu(self._inputs)
        if getattr(self, "_key", None) is None:
            self._update_key()  # update key when inputs are set

        tileables = []
        for j in range(output_limit):
            create_tensor_kw = kw.copy()
            if kws:
                create_tensor_kw.update(kws[j])
            tileable = self._create_tileable(j, **create_tensor_kw)
            tileables.append(tileable)

        self.outputs = tileables
        if len(tileables) > 1:
            # for each output tileable, hold the reference to the other outputs
            # so that either no one or everyone are gc collected
            for j, t in enumerate(tileables):
                t.data._siblings = [
                    tileable.data for tileable in tileables[:j] + tileables[j + 1 :]
                ]
        return tileables

    def new_tileables(
        self, inputs: List[TileableType], kws: List[dict] = None, **kw
    ) -> List[TileableType]:
        """
        Create tileable objects(Tensors or DataFrames).

        This is a base function for create tileable objects like tensors or dataframes,
        it will be called inside the `new_tensors` and `new_dataframes`.
        If eager mode is on, it will trigger the execution after tileable objects are created.

        Parameters
        ----------
        inputs : list
            Input tileables
        kws : List[dict]
            Kwargs for each output.
        kw : dict
            Common kwargs for all outputs.

        Returns
        -------
        tileables : list
            Output tileables.

        .. note::
            It's a final method, do not override.
            Override the method `_new_tileables` if needed.
        """
        tileables = self._new_tileables(inputs, kws=kws, **kw)
        if is_eager_mode():
            ExecutableTuple(tileables).execute()
        return tileables

    def new_tileable(
        self, inputs: List[TileableType], kws: List[Dict] = None, **kw
    ) -> TileableType:
        if getattr(self, "output_limit") != 1:
            raise TypeError("cannot new chunk with more than 1 outputs")

        return self.new_tileables(inputs, kws=kws, **kw)[0]

    @classmethod
    def pre_tile(cls, op: OperandType):
        """
        Operation before tile.

        Parameters
        ----------
        op : OperandType
          Operand to tile
        """

    @classmethod
    def tile(cls, op: OperandType):
        raise NotImplementedError

    @classmethod
    def post_tile(cls, op: OperandType, results: List[TileableType]):
        """
        Operation after tile.

        Parameters
        ----------
        op : OperandType
          Operand to tile.
        results: list
          List of tiled results.
        """

    @classmethod
    def pre_execute(cls, ctx: Union[dict, Context], op: OperandType):
        """
        Operation before execute.

        Parameters
        ----------
        ctx : dict
            Data store.
        op : OperandType
            Operand to execute.
        """

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: OperandType):
        raise NotImplementedError

    @classmethod
    def post_execute(cls, ctx: Union[dict, Context], op: OperandType):
        """
        Operand before execute.

        Parameters
        ----------
        ctx : dict
            Data store
        op : OperandType
            Operand to execute.
        """

    @classmethod
    def estimate_size(cls, ctx: dict, op: OperandType):
        from .fetch import FetchShuffle

        # when sizes of all outputs are deterministic, return directly
        outputs = op.outputs
        if all(
            not c.is_sparse() and hasattr(c, "nbytes") and not np.isnan(c.nbytes)
            for c in outputs
        ):
            for out in outputs:
                ctx[out.key] = (out.nbytes, out.nbytes)
            return

        pure_dep_keys = set(
            inp.key
            for inp, is_dep in zip(op.inputs or (), op.pure_depends or ())
            if is_dep
        )
        exec_sizes = [0]
        for inp in op.inputs or ():
            if inp.key in pure_dep_keys:
                continue
            try:
                if isinstance(inp.op, FetchShuffle):
                    keys_and_shapes = inp.extra_params.get("_shapes", dict()).items()
                else:
                    keys_and_shapes = [(inp.key, getattr(inp, "shape", None))]

                # execution size of a specific data chunk may be
                # larger than stored type due to objects
                for key, shape in keys_and_shapes:
                    exec_sizes.append(ctx[key][0])
            except KeyError:
                if not op.sparse:
                    inp_size = calc_data_size(inp)
                    if not np.isnan(inp_size):
                        exec_sizes.append(inp_size)
        if any(c.is_sparse() for c in op.inputs):
            exec_size = sum(exec_sizes)
        else:
            exec_size = max(exec_sizes)

        total_out_size = 0
        chunk_sizes = dict()
        for out in outputs:
            try:
                if not out.is_sparse():
                    chunk_size = calc_data_size(out)
                else:
                    chunk_size = exec_size
                if np.isnan(chunk_size):
                    raise TypeError
                chunk_sizes[out.key] = chunk_size
                total_out_size += chunk_size
            except (AttributeError, TypeError, ValueError):
                pass

        exec_size = max(exec_size, total_out_size)
        memory_scale = op.memory_scale or 1.0
        for out in outputs:
            if out.key in ctx:
                continue
            if out.key in chunk_sizes:
                result_size = chunk_sizes[out.key]
            else:
                result_size = max(
                    exec_size // len(outputs),
                    total_out_size // max(len(chunk_sizes), 1),
                )
            try:
                if getattr(out, "dtype", None) is not None and out.is_sparse():
                    max_sparse_size = (
                        out.nbytes
                        + np.dtype(np.int64).itemsize * np.prod(out.shape) * out.ndim
                    )
                else:
                    max_sparse_size = np.nan
            except TypeError:  # pragma: no cover
                max_sparse_size = np.nan
            if not np.isnan(max_sparse_size):
                result_size = min(result_size, max_sparse_size)
            ctx[out.key] = (result_size, int(exec_size * memory_scale // len(outputs)))

    @classmethod
    def concat_tileable_chunks(cls, tileable: TileableType):
        raise NotImplementedError

    @classmethod
    def create_tileable_from_chunks(
        cls, chunks: List[ChunkType], inputs: List[TileableType] = None, **kw
    ) -> TileableType:
        raise NotImplementedError

    def get_fetch_op_cls(self, obj: ChunkType):
        from .shuffle import ShuffleProxy

        output_types = get_output_types(obj, unknown_as=OutputType.object)
        fetch_cls, fetch_shuffle_cls = get_fetch_class(output_types[0])
        if isinstance(self, ShuffleProxy):
            cls = fetch_shuffle_cls
        else:
            cls = fetch_cls

        def _inner(**kw):
            return cls(output_types=output_types, **kw)

        return _inner

    def get_fuse_op_cls(self, obj: ChunkType):
        raise NotImplementedError

    @classmethod
    def register_executor(cls, executor: Callable):
        _op_type_to_executor[cls] = executor

    @classmethod
    def unregister_executor(cls):
        del _op_type_to_executor[cls]

    @classmethod
    def register_size_estimator(cls, size_estimator: Callable):
        _op_type_to_size_estimator[cls] = size_estimator

    @classmethod
    def unregister_size_estimator(cls):
        del _op_type_to_size_estimator[cls]


def execute(results: Dict[str, Any], op: OperandType):
    try:
        executor = _op_type_to_executor[type(op)]
    except KeyError:
        executor = type(op).execute

    # pre execute
    op.pre_execute(results, op)
    succeeded = False
    try:
        if UFuncTypeError is None:  # pragma: no cover
            return executor(results, op)
        else:
            # Cast `UFuncTypeError` to `TypeError` since subclasses of the former is unpickleable.
            # The `UFuncTypeError` was introduced by numpy#12593 since v1.17.0.
            try:
                result = executor(results, op)
                succeeded = True
                op_executed_number.record(1, {"op": op.__class__.__name__})
                return result
            except UFuncTypeError as e:  # pragma: no cover
                raise TypeError(str(e)).with_traceback(sys.exc_info()[2]) from None
    except NotImplementedError:
        for op_cls in type(op).__mro__:
            if op_cls in _op_type_to_executor:
                executor = _op_type_to_executor[op_cls]
                _op_type_to_executor[type(op)] = executor
                result = executor(results, op)
                succeeded = True
                return result
        raise KeyError(f"No handler found for op: {op}")
    finally:
        if succeeded:
            op.post_execute(results, op)


def estimate_size(results: Dict[str, Any], op: OperandType):
    try:
        size_estimator = _op_type_to_size_estimator[type(op)]
    except KeyError:
        size_estimator = type(op).estimate_size

    try:
        return size_estimator(results, op)
    except NotImplementedError:
        for op_cls in type(op).__mro__:
            if op_cls in _op_type_to_size_estimator:
                size_estimator = _op_type_to_size_estimator[op_cls]
                _op_type_to_size_estimator[type(op)] = size_estimator
                return size_estimator(results, op)
        raise KeyError(f"No handler found for op: {op} to estimate size")



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Iterable

import pandas as pd

from ...utils import adapt_mars_docstring
from .string_ import _string_method_to_handlers, SeriesStringMethod
from .datetimes import _datetime_method_to_handlers, SeriesDatetimeMethod


class StringAccessor:
    """
    Vectorized string functions for Series and Index.
    NAs stay NA unless handled otherwise by a particular method.
    Patterned after Python's string methods, with some inspiration from
    R's stringr package.
    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series(["A_Str_Series"])
    >>> s.execute()
    0    A_Str_Series
    dtype: object
    >>> s.str.split("_").execute()
    0    [A, Str, Series]
    dtype: object
    >>> s.str.replace("_", "").execute()
    0    AStrSeries
    dtype: object
    """

    def __init__(self, series):
        self._series = series

    @classmethod
    def _gen_func(cls, method):
        @wraps(getattr(pd.Series.str, method))
        def _inner(self, *args, **kwargs):
            op = SeriesStringMethod(
                method=method, method_args=args, method_kwargs=kwargs
            )
            return op(self._series)

        _inner.__doc__ = adapt_mars_docstring(getattr(pd.Series.str, method).__doc__)
        return _inner

    def __getitem__(self, item):
        return self._gen_func("__getitem__")(self, item)

    def __dir__(self) -> Iterable[str]:
        s = set(super().__dir__())
        s.update(_string_method_to_handlers.keys())
        return list(s)

    @classmethod
    def _register(cls, method):
        setattr(cls, method, cls._gen_func(method))

    def split(self, pat=None, n=-1, expand=False):
        r"""
        Split strings around given separator/delimiter.

        Splits the string in the Series/Index from the beginning,
        at the specified delimiter string. Equivalent to :meth:`str.split`.

        Parameters
        ----------
        pat : str, optional
            String or regular expression to split on.
            If not specified, split on whitespace.
        n : int, default -1 (all)
            Limit number of splits in output.
            ``None``, 0 and -1 will be interpreted as return all splits.
        expand : bool, default False
            Expand the splitted strings into separate columns.

            * If ``True``, return DataFrame/MultiIndex expanding dimensionality.
            * If ``False``, return Series/Index, containing lists of strings.

        Returns
        -------
        Series, Index, DataFrame or MultiIndex
            Type matches caller unless ``expand=True`` (see Notes).

        See Also
        --------
        Series.str.split : Split strings around given separator/delimiter.
        Series.str.rsplit : Splits string around given separator/delimiter,
            starting from the right.
        Series.str.join : Join lists contained as elements in the Series/Index
            with passed delimiter.
        str.split : Standard library version for split.
        str.rsplit : Standard library version for rsplit.

        Notes
        -----
        The handling of the `n` keyword depends on the number of found splits:

        - If found splits > `n`,  make first `n` splits only
        - If found splits <= `n`, make all splits
        - If for a certain row the number of found splits < `n`,
          append `None` for padding up to `n` if ``expand=True``

        If using ``expand=True``, Series and Index callers return DataFrame and
        MultiIndex objects, respectively.

        Examples
        --------
        >>> import numpy as np
        >>> import mars.dataframe as md
        >>> s = md.Series(["this is a regular sentence",
        >>>                "https://docs.python.org/3/tutorial/index.html",
        >>>                np.nan])
        >>> s.execute()
        0                       this is a regular sentence
        1    https://docs.python.org/3/tutorial/index.html
        2                                              NaN
        dtype: object

        In the default setting, the string is split by whitespace.

        >>> s.str.split().execute()
        0                   [this, is, a, regular, sentence]
        1    [https://docs.python.org/3/tutorial/index.html]
        2                                                NaN
        dtype: object

        Without the `n` parameter, the outputs of `rsplit` and `split`
        are identical.

        >>> s.str.rsplit().execute()
        0                   [this, is, a, regular, sentence]
        1    [https://docs.python.org/3/tutorial/index.html]
        2                                                NaN
        dtype: object

        The `n` parameter can be used to limit the number of splits on the
        delimiter. The outputs of `split` and `rsplit` are different.

        >>> s.str.split(n=2).execute()
        0                     [this, is, a regular sentence]
        1    [https://docs.python.org/3/tutorial/index.html]
        2                                                NaN
        dtype: object

        >>> s.str.rsplit(n=2).execute()
        0                     [this is a, regular, sentence]
        1    [https://docs.python.org/3/tutorial/index.html]
        2                                                NaN
        dtype: object

        The `pat` parameter can be used to split by other characters.

        >>> s.str.split(pat = "/").execute()
        0                         [this is a regular sentence]
        1    [https:, , docs.python.org, 3, tutorial, index...
        2                                                  NaN
        dtype: object

        When using ``expand=True``, the split elements will expand out into
        separate columns. If NaN is present, it is propagated throughout
        the columns during the split.

        >>> s.str.split(expand=True).execute()
                                                       0     1     2        3
        0                                           this    is     a  regular
        1  https://docs.python.org/3/tutorial/index.html  None  None     None
        2                                            NaN   NaN   NaN      NaN \
                     4
        0     sentence
        1         None
        2          NaN

        For slightly more complex use cases like splitting the html document name
        from a url, a combination of parameter settings can be used.

        >>> s.str.rsplit("/", n=1, expand=True).execute()
                                            0           1
        0          this is a regular sentence        None
        1  https://docs.python.org/3/tutorial  index.html
        2                                 NaN         NaN

        Remember to escape special characters when explicitly using regular
        expressions.

        >>> s = pd.Series(["1+1=2"])
        >>> s.str.split(r"\+|=", expand=True).execute()
             0    1    2
        0    1    1    2
        """
        return self._gen_func("split")(self, pat=pat, n=n, expand=expand)

    def rsplit(self, pat=None, n=-1, expand=False):
        return self._gen_func("rsplit")(self, pat=pat, n=n, expand=expand)

    def cat(self, others=None, sep=None, na_rep=None, join="left"):
        return self._gen_func("cat")(
            self, others=others, sep=sep, na_rep=na_rep, join=join
        )

    rsplit.__doc__ = adapt_mars_docstring(pd.Series.str.rsplit.__doc__)
    cat.__doc__ = adapt_mars_docstring(pd.Series.str.cat.__doc__)


class DatetimeAccessor:
    def __init__(self, series):
        self._series = series

    @classmethod
    def _gen_func(cls, method, is_property):
        @wraps(getattr(pd.Series.dt, method))
        def _inner(self, *args, **kwargs):
            op = SeriesDatetimeMethod(
                method=method,
                is_property=is_property,
                method_args=args,
                method_kwargs=kwargs,
            )
            return op(self._series)

        _inner.__doc__ = adapt_mars_docstring(getattr(pd.Series.dt, method).__doc__)
        return _inner

    @classmethod
    def _register(cls, method):
        is_property = not callable(getattr(pd.Series.dt, method))
        func = cls._gen_func(method, is_property)
        if is_property:
            func = property(func)
        setattr(cls, method, func)

    def __dir__(self) -> Iterable[str]:
        s = set(super().__dir__())
        s.update(_datetime_method_to_handlers.keys())
        return list(s)


class CachedAccessor:
    def __init__(self, name: str, accessor) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor
        if self._name not in obj._accessors:
            obj._accessors[self._name] = self._accessor(obj)
        return obj._accessors[self._name]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np

from ... import opcodes, tensor as mt
from ...core import OutputType, recursive_tile
from ...serialization.serializables import Int16Field, ReferenceField
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin


class LearnCountNonzero(LearnOperand, LearnOperandMixin):
    _op_code_ = opcodes.COUNT_NONZERO

    axis = Int16Field("axis")
    sample_weight = ReferenceField("sample_weight")

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.sample_weight is not None:
            self.sample_weight = inputs[-1]

    def __call__(self, x, sample_weight=None):
        self.sample_weight = sample_weight
        self._output_types = [
            OutputType.scalar if self.axis is None else OutputType.tensor
        ]
        dtype = np.dtype(int)
        inputs = [x]
        if sample_weight is not None:
            dtype = sample_weight.dtype
            inputs = [x, sample_weight]

        if self.axis is None:
            shape = ()
        else:
            shape = (x.shape[1 - self.axis],)

        return self.new_tileable(inputs, shape=shape, dtype=dtype)

    @classmethod
    def tile(cls, op: "LearnCountNonzero"):
        input_tensor = op.inputs[0]
        out_tensor = op.outputs[0]

        if op.sample_weight is not None:
            if has_unknown_shape(input_tensor):
                yield
            sample_weight = yield from recursive_tile(
                op.sample_weight.rechunk({0: input_tensor.nsplits[0]})
            )
        else:
            sample_weight = None

        chunks = []
        for input_chunk in input_tensor.chunks:
            if sample_weight is None:
                weight_chunk = None
            else:
                weight_chunk = sample_weight.cix[(input_chunk.index[0],)]

            new_op = op.copy().reset_key()
            new_op.sample_weight = weight_chunk

            inputs = [input_chunk] if not weight_chunk else [input_chunk, weight_chunk]
            if op.axis is None:
                shape = (1, 1)
            elif op.axis == 0:
                shape = (1, input_chunk.shape[1])
            else:
                shape = (input_chunk.shape[0], 1)
            chunks.append(
                new_op.new_chunk(
                    inputs, shape=shape, dtype=out_tensor.dtype, index=input_chunk.index
                )
            )

        new_op = op.copy().reset_key()
        if op.axis is None:
            nsplits = tuple((1,) * len(split) for split in input_tensor.nsplits)
            shape = tuple(len(split) for split in input_tensor.nsplits)
        elif op.axis == 0:
            nsplits = ((1,) * len(input_tensor.nsplits[0]), input_tensor.nsplits[1])
            shape = (len(input_tensor.nsplits[0]), input_tensor.shape[1])
        else:
            nsplits = (input_tensor.nsplits[0], (1,) * len(input_tensor.nsplits[1]))
            shape = (input_tensor.shape[0], len(input_tensor.nsplits[1]))

        tileable = new_op.new_tileable(
            out_tensor.inputs,
            chunks=chunks,
            nsplits=nsplits,
            shape=shape,
            dtype=out_tensor.dtype,
        )
        return [(yield from recursive_tile(mt.sum(tileable, axis=op.axis)))]

    @classmethod
    def execute(cls, ctx, op: "LearnCountNonzero"):
        axis = op.axis
        X = ctx[op.inputs[0].key]
        sample_weight = (
            ctx[op.sample_weight.key] if op.sample_weight is not None else None
        )

        # We rely here on the fact that np.diff(Y.indptr) for a CSR
        # will return the number of nonzero entries in each row.
        # A bincount over Y.indices will return the number of nonzeros
        # in each column. See ``csr_matrix.getnnz`` in scipy >= 0.14.
        if axis is None:
            if sample_weight is None:
                res = X.nnz
            else:
                res = np.dot(np.diff(X.indptr), sample_weight)
        elif axis == 1:
            out = np.diff(X.indptr)
            if sample_weight is None:
                # astype here is for consistency with axis=0 dtype
                res = out.astype("intp")
            else:
                res = out * sample_weight
        else:
            if sample_weight is None:
                res = np.bincount(X.indices, minlength=X.shape[1])
            else:
                weights = np.repeat(sample_weight, np.diff(X.indptr))
                res = np.bincount(X.indices, minlength=X.shape[1], weights=weights)
        if np.isscalar(res):
            res = np.array([res])
        out_shape = op.outputs[0].shape
        if any(np.isnan(s) for s in out_shape):
            new_shape = list(out_shape)
            for i, s in enumerate(out_shape):
                if np.isnan(s):
                    new_shape[i] = -1
            out_shape = tuple(new_shape)
        ctx[op.outputs[0].key] = res.reshape(out_shape)


def count_nonzero(X, axis: Optional[int] = None, sample_weight=None):
    """A variant of X.getnnz() with extension to weighting on axis 0

    Useful in efficiently calculating multilabel metrics.

    Parameters
    ----------
    X : CSR sparse matrix of shape (n_samples, n_labels)
        Input data.

    axis : None, 0 or 1
        The axis on which the data is aggregated.

    sample_weight : array-like of shape (n_samples,), default=None
        Weight for each row of X.
    """
    if axis == -1:
        axis = 1
    elif axis == -2:
        axis = 0
    if axis is not None and axis not in (0, 1):
        raise ValueError(f"Unsupported axis: {axis}")

    X = mt.asarray(X)
    if sample_weight is not None:
        sample_weight = mt.asarray(sample_weight)

    op = LearnCountNonzero(axis=axis)
    return op(X, sample_weight=sample_weight)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import BoolField
from ...typing import TileableType
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..core import TensorOrder


class TensorIsIn(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.ISIN

    assume_unique = BoolField("assume_unique")
    invert = BoolField("invert")

    def __call__(self, element, test_elements):
        self.dtype = np.dtype(bool)
        return self.new_tensor(
            [element, test_elements], shape=element.shape, order=TensorOrder.C_ORDER
        )

    @classmethod
    def tile(cls, op):
        from ..merge.stack import TensorStack
        from ..reduction import TensorAll, TensorAny

        ar1, ar2 = op.inputs
        invert = op.invert
        out = op.outputs[0]

        out_chunks = []
        for ar1_chunk in ar1.chunks:
            to_concat_chunks = []
            for ar2_chunk in ar2.chunks:
                chunk_op = op.copy().reset_key()
                out_chunk = chunk_op.new_chunk(
                    [ar1_chunk, ar2_chunk],
                    dtype=out.dtype,
                    shape=ar1_chunk.shape,
                    order=out.order,
                    index=ar1_chunk.index,
                )
                to_concat_chunks.append(out_chunk)
            if len(to_concat_chunks) == 1:
                out_chunks.append(to_concat_chunks[0])
            else:
                # concat chunks
                concat_op = TensorStack(axis=0)
                shape = (len(to_concat_chunks),) + ar1_chunk.shape
                concat_chunk = concat_op.new_chunk(
                    to_concat_chunks, shape=shape, dtype=out.dtype, order=out.order
                )
                if not invert:
                    chunk_op = TensorAny(axis=(0,), dtype=out.dtype)
                    out_chunk = chunk_op.new_chunk(
                        [concat_chunk],
                        shape=ar1_chunk.shape,
                        dtype=out.dtype,
                        order=out.order,
                        index=ar1_chunk.index,
                    )
                else:
                    chunk_op = TensorAll(axis=(0,), dtype=out.dtype)
                    out_chunk = chunk_op.new_chunk(
                        [concat_chunk],
                        shape=ar1_chunk.shape,
                        dtype=out.dtype,
                        order=out.order,
                        index=ar1_chunk.index,
                    )
                out_chunks.append(out_chunk)

        params = out.params.copy()
        params["nsplits"] = ar1.nsplits
        params["chunks"] = out_chunks
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        (element, test_elements), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            ctx[op.outputs[0].key] = xp.isin(
                element, test_elements, assume_unique=op.assume_unique, invert=op.invert
            )


def isin(
    element: Union[TileableType, np.ndarray],
    test_elements: Union[TileableType, np.ndarray, list],
    assume_unique: bool = False,
    invert: bool = False,
):
    """
    Calculates `element in test_elements`, broadcasting over `element` only.
    Returns a boolean array of the same shape as `element` that is True
    where an element of `element` is in `test_elements` and False otherwise.

    Parameters
    ----------
    element : array_like
        Input tensor.
    test_elements : array_like
        The values against which to test each value of `element`.
        This argument is flattened if it is a tensor or array_like.
        See notes for behavior with non-array-like parameters.
    assume_unique : bool, optional
        If True, the input tensors are both assumed to be unique, which
        can speed up the calculation.  Default is False.
    invert : bool, optional
        If True, the values in the returned tensor are inverted, as if
        calculating `element not in test_elements`. Default is False.
        ``mt.isin(a, b, invert=True)`` is equivalent to (but faster
        than) ``mt.invert(mt.isin(a, b))``.

    Returns
    -------
    isin : Tensor, bool
        Has the same shape as `element`. The values `element[isin]`
        are in `test_elements`.

    See Also
    --------
    in1d                  : Flattened version of this function.

    Notes
    -----

    `isin` is an element-wise function version of the python keyword `in`.
    ``isin(a, b)`` is roughly equivalent to
    ``mt.array([item in b for item in a])`` if `a` and `b` are 1-D sequences.

    `element` and `test_elements` are converted to tensors if they are not
    already. If `test_elements` is a set (or other non-sequence collection)
    it will be converted to an object tensor with one element, rather than a
    tensor of the values contained in `test_elements`. This is a consequence
    of the `tensor` constructor's way of handling non-sequence collections.
    Converting the set to a list usually gives the desired behavior.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> element = 2*mt.arange(4).reshape((2, 2))
    >>> element.execute()
    array([[0, 2],
           [4, 6]])
    >>> test_elements = [1, 2, 4, 8]
    >>> mask = mt.isin(element, test_elements)
    >>> mask.execute()
    array([[ False,  True],
           [ True,  False]])
    >>> element[mask].execute()
    array([2, 4])
    >>> mask = mt.isin(element, test_elements, invert=True)
    >>> mask.execute()
    array([[ True, False],
           [ False, True]])
    >>> element[mask]
    array([0, 6])

    Because of how `array` handles sets, the following does not
    work as expected:

    >>> test_set = {1, 2, 4, 8}
    >>> mt.isin(element, test_set).execute()
    array([[ False, False],
           [ False, False]])

    Casting the set to a list gives the expected result:

    >>> mt.isin(element, list(test_set)).execute()
    array([[ False,  True],
           [ True,  False]])
    """
    element, test_elements = astensor(element), astensor(test_elements).ravel()
    op = TensorIsIn(assume_unique=assume_unique, invert=invert)
    return op(element, test_elements)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import logging
import os
from numbers import Number
from typing import Dict

from ....serialization.ray import register_ray_serializers, unregister_ray_serializers
from ....utils import lazy_import
from ...driver import BaseActorDriver
from .utils import process_placement_to_address, addresses_to_placement_group_info

ray = lazy_import("ray")
logger = logging.getLogger(__name__)


class RayActorDriver(BaseActorDriver):
    _cluster_info = dict()

    @classmethod
    def setup_cluster(cls, address_to_resources: Dict[str, Dict[str, Number]]):
        logger.info("Setup cluster with %s", address_to_resources)
        # Note: Deep copy the dict to keep the origin values, because `bundles`
        # returned by `addresses_to_placement_group_info()` will be modified
        # by `ray.util.placement_group()`
        original_address_to_resources = copy.deepcopy(address_to_resources)
        pg_name, bundles = addresses_to_placement_group_info(address_to_resources)
        logger.info("Creating placement group %s with bundles %s.", pg_name, bundles)
        pg = ray.util.placement_group(name=pg_name, bundles=bundles, strategy="SPREAD")
        create_pg_timeout = 120
        done, _ = ray.wait([pg.ready()], timeout=create_pg_timeout)
        if not done:  # pragma: no cover
            raise Exception(
                f"""Can't create placement group {pg.bundle_specs} in {create_pg_timeout} seconds"""
            )
        cluster_info = {
            "original_address_to_resources": original_address_to_resources,
            "address_to_resources": address_to_resources,
            "pg_name": pg_name,
            "pg_group": pg,
            "main_pool_handles": [],  # Hold actor_handle to avoid actor being freed.
        }
        logger.info("Create placement group success.")
        cls._cluster_info = cluster_info
        register_ray_serializers()

    @classmethod
    def stop_cluster(cls):
        logger.info("Stopping cluster %s.", cls._cluster_info)
        if not cls._cluster_info:  # pragma: no cover
            return
        pg_name = cls._cluster_info["pg_name"]
        pg = cls._cluster_info["pg_group"]
        for index, bundle_spec in enumerate(pg.bundle_specs):
            # Main pool took a process.
            # If supervisor is created in the same node with worker, it will take a process too.
            n_process = int(bundle_spec["CPU"]) + 2
            for process_index in reversed(range(n_process)):
                address = process_placement_to_address(
                    pg_name, index, process_index=process_index
                )
                try:
                    if "COV_CORE_SOURCE" in os.environ:  # pragma: no cover
                        # must clean up first, or coverage info lost.
                        # must save the local reference until this is fixed:
                        # https://github.com/ray-project/ray/issues/7815
                        ray_actor = ray.get_actor(address)
                        ray.get(ray_actor.cleanup.remote())
                    ray.kill(ray.get_actor(address))
                except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                    pass
        ray.util.remove_placement_group(pg)
        cls._cluster_info = dict()
        unregister_ray_serializers()
        logger.info("Stopped cluster %s.", pg_name)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Dict
from ....resource import Resource


def get_band_resources_from_config(
    backend_execution_config: Dict,
) -> List[Dict[str, Resource]]:
    config = backend_execution_config
    n_worker: int = config["n_worker"]
    n_cpu: int = config["n_cpu"]
    mem_bytes: int = config["mem_bytes"]
    cuda_devices: List[List[int]] = config["cuda_devices"]

    bands_to_resource = []
    worker_cpus = n_cpu // n_worker
    cuda_devices = cuda_devices or ([[]] * n_worker)
    if sum(len(devices) for devices in cuda_devices) == 0:
        assert worker_cpus > 0, (
            f"{n_cpu} cpus are not enough " f"for {n_worker}, try to decrease workers."
        )
    mem_bytes = mem_bytes // n_worker
    for _, devices in zip(range(n_worker), cuda_devices):
        worker_band_to_resource = dict()
        worker_band_to_resource["numa-0"] = Resource(
            num_cpus=worker_cpus, mem_bytes=mem_bytes
        )
        for i in devices:  # pragma: no cover
            worker_band_to_resource[f"gpu-{i}"] = Resource(num_gpus=1)
        bands_to_resource.append(worker_band_to_resource)
    return bands_to_resource



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, StringField, AnyField
from ..array_utils import as_same_device, device, get_array_module
from ..utils import get_order
from .core import TensorHasInput
from .array import tensor


class SparseToDense(TensorHasInput):
    _op_type_ = OperandDef.SPARSE_TO_DENSE

    _input = KeyField("input")
    _order = StringField("order")
    _fill_value = AnyField("fill_value")

    def __init__(self, fill_value=None, order=None, **kw):
        super().__init__(_fill_value=fill_value, sparse=False, _order=order, **kw)

    @property
    def order(self):
        return self._order

    @property
    def fill_value(self):
        return self._fill_value

    @classmethod
    def execute(cls, ctx, op):
        fill_value = op.fill_value
        out = op.outputs[0]
        (inp,), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            if fill_value is None:
                ctx[out.key] = inp.toarray().astype(
                    out.dtype, order=op.order, copy=False
                )
            else:
                xp = get_array_module(xp)
                spmatrix = inp.spmatrix
                inds = spmatrix.nonzero()
                ret = xp.full(inp.shape, fill_value, dtype=out.dtype, order=op.order)
                ret[inds] = spmatrix.data
                ctx[out.key] = ret


def fromsparse(a, order="C", fill_value=None):
    a = tensor(a)
    if not a.issparse():
        return a.astype(a.dtype, order=order, copy=False)

    tensor_order = get_order(
        order,
        None,
        available_options="CF",
        err_msg="only 'C' or 'F' order is permitted",
    )
    op = SparseToDense(dtype=a.dtype, gpu=a.op.gpu, order=order, fill_value=fill_value)
    return op(a, order=tensor_order)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import sys
from collections import defaultdict
from typing import Dict, List, Optional, Set

from ..... import oscar as mo
from .....core import ChunkGraph, TileContext
from .....core.context import set_context
from .....core.operand import (
    Fetch,
    MapReduceOperand,
    OperandStage,
    ShuffleProxy,
)
from .....lib.aio import alru_cache
from .....oscar.profiling import (
    ProfilingData,
)
from .....resource import Resource
from .....typing import TileableType, BandType
from .....utils import Timer
from ....context import ThreadedServiceContext
from ....cluster.api import ClusterAPI
from ....lifecycle.api import LifecycleAPI
from ....meta.api import MetaAPI
from ....scheduling import SchedulingAPI
from ....subtask import Subtask, SubtaskResult, SubtaskStatus, SubtaskGraph
from ...core import Task
from ..api import TaskExecutor, register_executor_cls
from .config import MarsExecutionConfig
from .resource import ResourceEvaluator
from .stage import TaskStageProcessor

logger = logging.getLogger(__name__)


def _get_n_reducers(subtask: Subtask) -> int:
    return len(
        [
            r
            for r in subtask.chunk_graph
            if isinstance(r.op, MapReduceOperand) and r.op.stage == OperandStage.reduce
        ]
    )


@register_executor_cls
class MarsTaskExecutor(TaskExecutor):
    name = "mars"
    _stage_processors: List[TaskStageProcessor]
    _stage_tile_progresses: List[float]
    _cur_stage_processor: Optional[TaskStageProcessor]
    _meta_updated_tileables: Set[TileableType]

    def __init__(
        self,
        config: MarsExecutionConfig,
        task: Task,
        tile_context: TileContext,
        cluster_api: ClusterAPI,
        lifecycle_api: LifecycleAPI,
        scheduling_api: SchedulingAPI,
        meta_api: MetaAPI,
        resource_evaluator: ResourceEvaluator,
    ):
        self._config = config
        self._task = task
        self._tileable_graph = task.tileable_graph
        self._raw_tile_context = tile_context.copy()
        self._tile_context = tile_context
        self._session_id = task.session_id

        # api
        self._cluster_api = cluster_api
        self._lifecycle_api = lifecycle_api
        self._scheduling_api = scheduling_api
        self._meta_api = meta_api

        self._stage_processors = []
        self._stage_tile_progresses = []
        self._cur_stage_processor = None
        self._lifecycle_processed_tileables = set()
        self._subtask_decref_events = dict()
        self._meta_updated_tileables = set()

        # Evaluate and initialize subtasks required resource.
        self._resource_evaluator = resource_evaluator

    @classmethod
    async def create(
        cls,
        config: MarsExecutionConfig,
        *,
        session_id: str,
        address: str,
        task: Task,
        tile_context: TileContext,
        **kwargs,
    ) -> "MarsTaskExecutor":
        assert (
            len(kwargs) == 0
        ), f"Unexpected kwargs for {cls.__name__}.create: {kwargs}"
        cluster_api, lifecycle_api, scheduling_api, meta_api = await cls._get_apis(
            session_id, address
        )
        resource_evaluator = await ResourceEvaluator.create(
            config.get_execution_config(),
            session_id=task.session_id,
            task_id=task.task_id,
            cluster_api=cluster_api,
        )
        await cls._init_context(session_id, address)
        return cls(
            config,
            task,
            tile_context,
            cluster_api,
            lifecycle_api,
            scheduling_api,
            meta_api,
            resource_evaluator,
        )

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def _get_apis(cls, session_id: str, address: str):
        return await asyncio.gather(
            ClusterAPI.create(address),
            LifecycleAPI.create(session_id, address),
            SchedulingAPI.create(session_id, address),
            MetaAPI.create(session_id, address),
        )

    @classmethod
    async def _init_context(cls, session_id: str, address: str):
        loop = asyncio.get_running_loop()
        context = ThreadedServiceContext(
            session_id, address, address, address, loop=loop
        )
        await context.init()
        set_context(context)

    async def __aenter__(self):
        profiling = ProfilingData[self._task.task_id, "general"]
        # incref fetch tileables to ensure fetch data not deleted
        with Timer() as timer:
            await self._incref_fetch_tileables()
        profiling.set("incref_fetch_tileables", timer.duration)

    async def execute_subtask_graph(
        self,
        stage_id: str,
        subtask_graph: SubtaskGraph,
        chunk_graph: ChunkGraph,
        tile_context: TileContext,
        context=None,
    ):
        available_bands = await self.get_available_band_resources()
        await self._incref_result_tileables()
        stage_processor = TaskStageProcessor(
            stage_id,
            self._task,
            chunk_graph,
            subtask_graph,
            list(available_bands),
            tile_context,
            self._scheduling_api,
            self._meta_api,
        )
        await self._incref_stage(stage_processor)
        await self._resource_evaluator.evaluate(stage_processor)
        self._stage_processors.append(stage_processor)
        self._cur_stage_processor = stage_processor
        # get the tiled progress for current stage
        prev_progress = sum(self._stage_tile_progresses)
        curr_tile_progress = self._tile_context.get_all_progress() - prev_progress
        self._stage_tile_progresses.append(curr_tile_progress)
        return await stage_processor.run()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # clean ups
        decrefs = []
        error_or_cancelled = False
        for stage_processor in self._stage_processors:
            if stage_processor.error_or_cancelled():
                error_or_cancelled = True
            decrefs.append(self._decref_stage.delay(stage_processor))
        await self._decref_stage.batch(*decrefs)
        # revert fetch incref
        await self._decref_fetch_tileables()
        if error_or_cancelled:
            # revert result incref if error or cancelled
            await self._decref_result_tileables()
        await self._resource_evaluator.report()

    async def get_available_band_resources(self) -> Dict[BandType, Resource]:
        async for bands in self._cluster_api.watch_all_bands():
            if bands:
                return bands

    async def get_progress(self) -> float:
        # get progress of stages
        executor_progress = 0.0
        assert len(self._stage_tile_progresses) == len(self._stage_processors)
        for stage_processor, stage_tile_progress in zip(
            self._stage_processors, self._stage_tile_progresses
        ):
            if stage_processor.subtask_graph is None:  # pragma: no cover
                # generating subtask
                continue
            n_subtask = len(stage_processor.subtask_graph)
            if n_subtask == 0:  # pragma: no cover
                continue
            progress = sum(
                result.progress for result in stage_processor.subtask_results.values()
            )
            progress += sum(
                result.progress
                for subtask_key, result in stage_processor.subtask_snapshots.items()
                if subtask_key not in stage_processor.subtask_results
            )
            subtask_progress = progress / n_subtask
            executor_progress += subtask_progress * stage_tile_progress
        return executor_progress

    async def cancel(self):
        if self._cur_stage_processor is not None:
            await self._cur_stage_processor.cancel()

    async def set_subtask_result(self, subtask_result: SubtaskResult):
        if self._cur_stage_processor is None or (
            subtask_result.stage_id
            and self._cur_stage_processor.stage_id != subtask_result.stage_id
        ):
            logger.warning(
                "Stage %s for subtask %s not exists, got stale subtask result %s which may be "
                "speculative execution from previous stages, just ignore it.",
                subtask_result.stage_id,
                subtask_result.subtask_id,
                subtask_result,
            )
            return
        stage_processor = self._cur_stage_processor
        subtask = stage_processor.subtask_id_to_subtask[subtask_result.subtask_id]

        prev_result = stage_processor.subtask_results.get(subtask)
        if prev_result and (
            prev_result.status == SubtaskStatus.succeeded
            or prev_result.progress > subtask_result.progress
        ):
            logger.info(
                "Skip set subtask %s with result %s, previous result is %s.",
                subtask.subtask_id,
                subtask_result,
                prev_result,
            )
            # For duplicate run of subtasks, if the progress is smaller or the subtask has finished or canceled
            # in task speculation, just do nothing.
            # TODO(chaokunyang) If duplicate run of subtasks failed, it may be the fault in worker node,
            #  print the exception, and if multiple failures on the same node, remove the node from the cluster.
            return
        if subtask_result.bands:
            [band] = subtask_result.bands
        else:
            band = None
        stage_processor.subtask_snapshots[subtask] = subtask_result.update(
            stage_processor.subtask_snapshots.get(subtask)
        )
        if subtask_result.status.is_done:
            # update stage_processor.subtask_results to avoid concurrent set_subtask_result
            # since we release lock when `_decref_input_subtasks`.
            stage_processor.subtask_results[subtask] = subtask_result.update(
                stage_processor.subtask_results.get(subtask)
            )
            try:
                # Since every worker will call supervisor to set subtask result,
                # we need to release actor lock to make `decref_chunks` parallel to avoid blocking
                # other `set_subtask_result` calls.
                # If speculative execution enabled, concurrent subtasks may got error since input chunks may
                # got deleted. But it's OK because the current subtask run has succeed.
                if subtask.subtask_id not in stage_processor.decref_subtask:
                    stage_processor.decref_subtask.add(subtask.subtask_id)
                    await self._decref_input_subtasks(
                        subtask, stage_processor.subtask_graph
                    )

            except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
                logger.debug(
                    "Decref input subtasks for subtask %s failed.", subtask.subtask_id
                )
                _, err, tb = sys.exc_info()
                if subtask_result.status not in (
                    SubtaskStatus.errored,
                    SubtaskStatus.cancelled,
                ):
                    subtask_result.status = SubtaskStatus.errored
                    subtask_result.error = err
                    subtask_result.traceback = tb
            await stage_processor.set_subtask_result(subtask_result, band=band)

    def get_stage_processors(self):
        return self._stage_processors

    async def _incref_fetch_tileables(self):
        # incref fetch tileables in tileable graph to prevent them from deleting
        to_incref_tileable_keys = [
            tileable.op.source_key
            for tileable in self._tileable_graph
            if isinstance(tileable.op, Fetch) and tileable in self._raw_tile_context
        ]
        await self._lifecycle_api.incref_tileables(to_incref_tileable_keys)

    async def _decref_fetch_tileables(self):
        fetch_tileable_keys = [
            tileable.op.source_key
            for tileable in self._tileable_graph
            if isinstance(tileable.op, Fetch) and tileable in self._raw_tile_context
        ]
        await self._lifecycle_api.decref_tileables(fetch_tileable_keys)

    def _get_tiled(self, tileable: TileableType):
        tileable = tileable.data if hasattr(tileable, "data") else tileable
        return self._tile_context[tileable]

    async def _incref_result_tileables(self):
        processed = self._lifecycle_processed_tileables
        # track and incref result tileables if tiled
        tracks = [], []
        for result_tileable in self._tileable_graph.result_tileables:
            if result_tileable in processed:  # pragma: no cover
                continue
            try:
                tiled_tileable = self._get_tiled(result_tileable)
                tracks[0].append(result_tileable.key)
                tracks[1].append(
                    self._lifecycle_api.track.delay(
                        result_tileable.key, [c.key for c in tiled_tileable.chunks]
                    )
                )
                processed.add(result_tileable)
            except KeyError:
                # not tiled, skip
                pass
        if tracks:
            await self._lifecycle_api.track.batch(*tracks[1])
            await self._lifecycle_api.incref_tileables(tracks[0])

    async def _decref_result_tileables(self):
        await self._lifecycle_api.decref_tileables(
            [t.key for t in self._lifecycle_processed_tileables]
        )

    async def _incref_stage(self, stage_processor: "TaskStageProcessor"):
        subtask_graph = stage_processor.subtask_graph
        incref_chunk_key_to_counts = defaultdict(lambda: 0)
        for subtask in subtask_graph:
            # for subtask has successors, incref number of successors
            n = subtask_graph.count_successors(subtask)
            for c in subtask.chunk_graph.results:
                incref_chunk_key_to_counts[c.key] += n
            # process reducer, incref mapper chunks
            for pre_graph in subtask_graph.iter_predecessors(subtask):
                for chk in pre_graph.chunk_graph.results:
                    if isinstance(chk.op, ShuffleProxy):
                        n_reducers = _get_n_reducers(subtask)
                        for map_chunk in chk.inputs:
                            incref_chunk_key_to_counts[map_chunk.key] += n_reducers
        result_chunks = stage_processor.chunk_graph.result_chunks
        for c in result_chunks:
            incref_chunk_key_to_counts[c.key] += 1
        logger.debug(
            "Incref chunks for stage %s: %s",
            stage_processor.stage_id,
            incref_chunk_key_to_counts,
        )
        await self._lifecycle_api.incref_chunks(
            list(incref_chunk_key_to_counts),
            counts=list(incref_chunk_key_to_counts.values()),
        )

    @classmethod
    def _get_decref_stage_chunk_key_to_counts(
        cls, stage_processor: "TaskStageProcessor"
    ) -> Dict[str, int]:
        decref_chunk_key_to_counts = defaultdict(lambda: 0)
        error_or_cancelled = stage_processor.error_or_cancelled()
        if stage_processor.subtask_graph:
            subtask_graph = stage_processor.subtask_graph
            if error_or_cancelled:
                # error or cancel, rollback incref for subtask results
                for subtask in subtask_graph:
                    if subtask.subtask_id in stage_processor.decref_subtask:
                        continue
                    stage_processor.decref_subtask.add(subtask.subtask_id)
                    # if subtask not executed, rollback incref of predecessors
                    for inp_subtask in subtask_graph.predecessors(subtask):
                        for c in inp_subtask.chunk_graph.results:
                            decref_chunk_key_to_counts[c.key] += 1
        # decref result of chunk graphs
        for c in stage_processor.chunk_graph.results:
            decref_chunk_key_to_counts[c.key] += 1
        return decref_chunk_key_to_counts

    @mo.extensible
    async def _decref_stage(self, stage_processor: "TaskStageProcessor"):
        decref_chunk_key_to_counts = self._get_decref_stage_chunk_key_to_counts(
            stage_processor
        )
        logger.debug(
            "Decref chunks when stage %s finish: %s",
            stage_processor.stage_id,
            decref_chunk_key_to_counts,
        )
        await self._lifecycle_api.decref_chunks(
            list(decref_chunk_key_to_counts),
            counts=list(decref_chunk_key_to_counts.values()),
        )

    @_decref_stage.batch
    async def _decref_stage(self, args_list, kwargs_list):
        decref_chunk_key_to_counts = defaultdict(lambda: 0)
        for args, kwargs in zip(args_list, kwargs_list):
            chunk_key_to_counts = self._get_decref_stage_chunk_key_to_counts(
                *args, **kwargs
            )
            for k, c in chunk_key_to_counts.items():
                decref_chunk_key_to_counts[k] += c
        logger.debug("Decref chunks when stages finish: %s", decref_chunk_key_to_counts)
        await self._lifecycle_api.decref_chunks(
            list(decref_chunk_key_to_counts),
            counts=list(decref_chunk_key_to_counts.values()),
        )

    async def _decref_input_subtasks(
        self, subtask: Subtask, subtask_graph: SubtaskGraph
    ):
        # make sure subtasks are decreffed only once
        if subtask.subtask_id not in self._subtask_decref_events:
            self._subtask_decref_events[subtask.subtask_id] = asyncio.Event()
        else:  # pragma: no cover
            await self._subtask_decref_events[subtask.subtask_id].wait()
            return

        decref_chunk_key_to_counts = defaultdict(lambda: 0)
        for in_subtask in subtask_graph.iter_predecessors(subtask):
            for result_chunk in in_subtask.chunk_graph.results:
                # for reducer chunk, decref mapper chunks
                if isinstance(result_chunk.op, ShuffleProxy):
                    n_reducers = _get_n_reducers(subtask)
                    for inp in result_chunk.inputs:
                        decref_chunk_key_to_counts[inp.key] += n_reducers
                decref_chunk_key_to_counts[result_chunk.key] += 1
        logger.debug(
            "Decref chunks %s when subtask %s finish",
            decref_chunk_key_to_counts,
            subtask.subtask_id,
        )
        await self._lifecycle_api.decref_chunks(
            list(decref_chunk_key_to_counts),
            counts=list(decref_chunk_key_to_counts.values()),
        )

        # `set_subtask_result` will be called when subtask finished
        # but report progress will call set_subtask_result too,
        # so it have risk to duplicate decrease some subtask input object reference,
        # it will cause object reference count lower zero
        # TODO(Catch-Bull): Pop asyncio.Event when current subtask `set_subtask_result`
        # will never be called
        self._subtask_decref_events[subtask.subtask_id].set()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from collections import defaultdict
from typing import Dict, Iterable, List, Tuple

from ....utils import lazy_import

ray = lazy_import("ray")
parallel_it = lazy_import("ray.util.iter", rename="parallel_it")
ml_dataset = lazy_import("ray.util.data", rename="ml_dataset")


class ChunkRefBatch:
    def __init__(self, shard_id: int, obj_refs: "ray.ObjectRef"):
        """Iterable batch holding a list of ray.ObjectRefs.

        Args:
            shard_id (int): id of the shard
            prefix (str): prefix name of the batch
            obj_refs (List[ray.ObjectRefs]): list of ray.ObjectRefs
        """
        self._shard_id = shard_id
        self._obj_refs = obj_refs

    @property
    def shard_id(self) -> int:
        return self._shard_id

    def __iter__(self) -> Iterable[pd.DataFrame]:
        """Returns the item_generator required from ParallelIteratorWorker."""
        for obj_ref in self._obj_refs:
            yield ray.get(obj_ref)


def _group_chunk_refs(
    chunk_addr_refs: List[Tuple[Tuple, "ray.ObjectRef"]], num_shards: int
):
    """Group fetched ray.ObjectRefs into a dict for later use.

    Args:
        chunk_addr_refs (List[Tuple[Tuple, ray.ObjectRef]]): a list of tuples of
            band & ray.ObjectRef of each chunk.
        num_shards (int): the number of shards that will be created for the MLDataset.

    Returns:
        Dict[str, List[ray.ObjectRef]]: a dict that defines which group of ray.ObjectRefs will
            be in an ChunkRefBatch.
    """
    group_to_obj_refs = defaultdict(list)
    if not num_shards:
        for addr, obj_ref in chunk_addr_refs:
            group_to_obj_refs[addr].append(obj_ref)
    else:
        splits = np.array_split([ref for _, ref in chunk_addr_refs], num_shards)
        for idx, split in enumerate(splits):
            group_to_obj_refs["group-" + str(idx)] = list(split)
    return group_to_obj_refs


def _rechunk_if_needed(df, num_shards: int = None):
    try:
        if num_shards:
            assert isinstance(num_shards, int) and num_shards > 0
            df = df.rebalance(axis=0, num_partitions=num_shards)
        df = df.rechunk({1: df.shape[1]})
        df = df.reset_index(drop=True)
        return df.execute()
    except Exception as e:  # pragma: no cover
        raise Exception(f"rechunk failed df.shape {df.shape}") from e


def to_ray_mldataset(df, num_shards: int = None):
    """Create a MLDataset from Mars DataFrame

    Args:
        df (mars.dataframe.Dataframe): the Mars DataFrame
        num_shards (int, optional): the number of shards that will be created
            for the MLDataset. Defaults to None.
            If num_shards is None, chunks will be grouped by nodes where they lie.
            Otherwise, chunks will be grouped by their order in DataFrame.

    Returns:
        a MLDataset
    """
    df = _rechunk_if_needed(df, num_shards)
    # chunk_addr_refs is fetched directly rather than in batches
    # during `fetch` procedure, it'll be checked that df has been executed
    # items in chunk_addr_refs are ordered by positions in df
    # while adjacent chunks may belong to different addrs, i.e.
    #       chunk1 for addr1,
    #       chunk2 & chunk3 for addr2,
    #       chunk4 for addr1
    fetched_infos: Dict[str, List] = df.fetch_infos(fields=["bands", "object_refs"])
    chunk_addr_refs: List[Tuple[Tuple, "ray.ObjectRef"]] = [
        (bands[0], object_refs[0])
        for bands, object_refs in zip(
            fetched_infos["bands"], fetched_infos["object_refs"]
        )
    ]
    group_to_obj_refs: Dict[str, List[ray.ObjectRef]] = _group_chunk_refs(
        chunk_addr_refs, num_shards
    )

    record_batches = []
    for rank, obj_refs in enumerate(group_to_obj_refs.values()):
        record_batches.append(ChunkRefBatch(shard_id=rank, obj_refs=obj_refs))
    worker_cls = ray.remote(num_cpus=0)(parallel_it.ParallelIteratorWorker)
    actors = [worker_cls.remote(g, False) for g in record_batches]
    it = parallel_it.from_actors(actors, "from_mars")
    dataset = ml_dataset.from_parallel_iter(it, need_convert=False, batch_size=0)
    # Hold mars dataframe to avoid mars dataframe and ray object gc.
    dataset.dataframe = df

    def __getstate__():
        state = dataset.__dict__.copy()
        state.pop("dataframe", None)
        return state

    # `dataframe` is not serializable by ray.
    dataset.__getstate__ = __getstate__
    return dataset



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import numpy as np

from .... import opcodes
from .... import options
from ....core import recursive_tile
from ....core.operand import OperandStage
from ....serialization.serializables import (
    KeyField,
    BoolField,
    DictField,
    Int64Field,
    AnyField,
)
from ....tensor.core import TensorOrder
from ....tensor.merge import TensorConcatenate
from ....tensor.array_utils import as_same_device, device, get_array_module
from ....utils import has_unknown_shape, ensure_own_data
from ...utils import gen_batches, get_chunk_n_rows
from ...utils.validation import _num_samples
from .core import PairwiseDistances


def _precompute_metric_params(X, Y, xp, metric=None, **kwds):  # pragma: no cover
    """Precompute data-derived metric parameters if not provided"""
    if metric == "seuclidean" and "V" not in kwds:
        if X is Y:
            V = xp.var(X, axis=0, ddof=1)
        else:
            V = xp.var(xp.vstack([X, Y]), axis=0, ddof=1)
        return {"V": V}
    if metric == "mahalanobis" and "VI" not in kwds:
        if X is Y:
            VI = xp.linalg.inv(xp.cov(X.T)).T
        else:
            VI = xp.linalg.inv(xp.cov(xp.vstack([X, Y]).T)).T
        return {"VI": VI}
    return {}


def _check_chunk_size(reduced, chunk_size):  # pragma: no cover
    """Checks chunk is a sequence of expected size or a tuple of same"""
    if reduced is None:
        return
    is_tuple = isinstance(reduced, tuple)
    if not is_tuple:
        reduced = (reduced,)
    if any(isinstance(r, tuple) or not hasattr(r, "__iter__") for r in reduced):
        raise TypeError(
            "reduce_func returned %r. "
            "Expected sequence(s) of length %d."
            % (reduced if is_tuple else reduced[0], chunk_size)
        )
    if any(_num_samples(r) != chunk_size for r in reduced):
        actual_size = tuple(_num_samples(r) for r in reduced)
        raise ValueError(
            "reduce_func returned object of length %s. "
            "Expected same length as input: %d."
            % (actual_size if is_tuple else actual_size[0], chunk_size)
        )


def _pariwise_distance_chunked(
    X, Y, reduce_func=None, metric="euclidean", working_memory=None, xp=None, **kwds
):
    if xp is np:
        from sklearn.metrics import pairwise_distances
    else:  # pragma: no cover
        from cuml.metrics import pairwise_distances

    n_samples_X = _num_samples(X)
    if metric == "precomputed":  # pragma: no cover
        slices = (slice(0, n_samples_X),)
    else:
        # We get as many rows as possible within our working_memory budget to
        # store len(Y) distances in each row of output.
        #
        # Note:
        #  - this will get at least 1 row, even if 1 row of distances will
        #    exceed working_memory.
        #  - this does not account for any temporary memory usage while
        #    calculating distances (e.g. difference of vectors in manhattan
        #    distance.
        chunk_n_rows = get_chunk_n_rows(
            row_bytes=8 * _num_samples(Y),
            max_n_rows=n_samples_X,
            working_memory=working_memory,
        )
        slices = gen_batches(n_samples_X, chunk_n_rows)

    # precompute data-derived metric params
    params = _precompute_metric_params(X, Y, xp, metric=metric, **kwds)
    kwds.update(**params)

    for sl in slices:
        if sl.start == 0 and sl.stop == n_samples_X:
            X_chunk = X  # enable optimised paths for X is Y
        else:
            X_chunk = X[sl]
        # call pairwise op's execute method to get the result
        D_chunk = pairwise_distances(
            ensure_own_data(X_chunk), ensure_own_data(Y), metric=metric, **kwds
        )
        if (X is Y or Y is None) and metric == "euclidean":
            # zeroing diagonal, taking care of aliases of "euclidean",
            # i.e. "l2"
            D_chunk.flat[
                sl.start :: _num_samples(X) + 1
            ] = 0  # pylint: disable=invalid-slice-index
        if reduce_func is not None:
            chunk_size = D_chunk.shape[0]
            D_chunk = reduce_func(D_chunk, sl.start)
            _check_chunk_size(D_chunk, chunk_size)
        yield D_chunk


class PairwiseDistancesTopk(PairwiseDistances):
    _op_type_ = opcodes.PAIRWISE_DISTANCES_TOPK

    _x = KeyField("x")
    _y = KeyField("y")
    _k = Int64Field("k")
    _metric = AnyField("metric")
    _metric_kwargs = DictField("metric_kwargs")
    _return_index = BoolField("return_index")
    _working_memory = AnyField("working_memory")
    # for chunks
    _y_offset = Int64Field("y_offset")

    def __init__(
        self,
        x=None,
        y=None,
        k=None,
        metric=None,
        metric_kwargs=None,
        return_index=None,
        working_memory=None,
        **kw
    ):
        super().__init__(
            _x=x,
            _y=y,
            _k=k,
            _metric=metric,
            _metric_kwargs=metric_kwargs,
            _return_index=return_index,
            _working_memory=working_memory,
            **kw
        )

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def k(self):
        return self._k

    @property
    def metric(self):
        return self._metric

    @property
    def metric_kwargs(self):
        return self._metric_kwargs

    @property
    def return_index(self):
        return self._return_index

    @property
    def working_memory(self):
        return self._working_memory

    @property
    def y_offset(self):
        return self._y_offset

    @property
    def output_limit(self):
        return 1 if not self._return_index or self.stage == OperandStage.map else 2

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.stage != OperandStage.agg:
            self._x = self._inputs[0]
            self._y = self._inputs[1]
        else:
            self._x = self._y = None

    def __call__(self, X, Y):
        from .pairwise import pairwise_distances

        # leverage pairwise_distances for checks
        d = pairwise_distances(X, Y, metric=self._metric, **self._metric_kwargs)

        if self._k > Y.shape[0]:
            self._k = Y.shape[0]

        X, Y = d.op.inputs

        shape_list = [X.shape[0]]
        shape_list.append(min(Y.shape[0], self._k))
        shape = tuple(shape_list)
        kws = [
            {
                "shape": shape,
                "order": TensorOrder.C_ORDER,
                "dtype": np.dtype(np.float64),
                "_type_": "distance",
            },
        ]
        if self._return_index:
            kws.append(
                {
                    "shape": shape,
                    "order": TensorOrder.C_ORDER,
                    "dtype": np.dtype(np.int64),
                    "_type_": "index",
                }
            )
            return self.new_tensors([X, Y], kws=kws)
        else:
            return self.new_tensors([X, Y], kws=kws)[0]

    @classmethod
    def _gen_out_chunks(cls, x_chunk, y_chunk, chunk_op):
        k = chunk_op.k
        i, j = x_chunk.index[0], y_chunk.index[0]

        distance_chunk_params = {
            "shape": (x_chunk.shape[0], k),
            "order": TensorOrder.C_ORDER,
            "dtype": np.dtype(np.float64),
            "index": (i, j),
            "_type_": "distance",
        }
        if chunk_op.return_index:
            index_chunk_params = {
                "shape": (x_chunk.shape[0], k),
                "order": TensorOrder.C_ORDER,
                "dtype": np.dtype(np.int64),
                "index": (i, j),
                "_type_": "index",
            }
            distance_chunk, index_chunk = chunk_op.new_chunks(
                [x_chunk, y_chunk], kws=[distance_chunk_params, index_chunk_params]
            )
            return distance_chunk, index_chunk
        else:
            return chunk_op.new_chunks([x_chunk, y_chunk], kws=[distance_chunk_params])[
                0
            ]

    @classmethod
    def tile(cls, op: "PairwiseDistancesTopk"):
        X, Y = op.x, op.y
        k = op.k

        if X.chunk_shape[1] > 1:
            X = yield from recursive_tile(X.rechunk({1: X.shape[1]}))

        if has_unknown_shape(Y):
            yield
        if Y.chunk_shape[1] > 1:
            Y = yield from recursive_tile(Y.rechunk({1: Y.shape[1]}))

        out_distance_chunks, out_index_chunks = [], []
        y_acc_chunk_shapes = [0] + np.cumsum(Y.nsplits[0]).tolist()
        for i in range(len(range(X.chunk_shape[0]))):
            x_chunk = X.cix[i, 0]
            y_chunk_shape = Y.chunk_shape[0]

            if y_chunk_shape == 1:
                chunk_op = op.copy().reset_key()
                y_chunk = Y.chunks[0]
                o = cls._gen_out_chunks(x_chunk, y_chunk, chunk_op)
                if chunk_op.return_index:
                    out_distance_chunks.append(o[0])
                    out_index_chunks.append(o[1])
                else:
                    out_distance_chunks.append(o)
            else:
                to_concat_chunks = []
                concat_size = 0
                for j in range(y_chunk_shape):
                    y_chunk = Y.cix[j, 0]
                    chunk_op = op.copy().reset_key()
                    chunk_op._y_offset = y_acc_chunk_shapes[j]
                    chunk_op.stage = OperandStage.map
                    size = min(k, y_chunk.shape[0])
                    o = chunk_op.new_chunk(
                        [x_chunk, y_chunk],
                        shape=(x_chunk.shape[0], size),
                        order=TensorOrder.C_ORDER,
                        index=(i, j),
                    )
                    to_concat_chunks.append(o)
                    concat_size += size

                concat_op = TensorConcatenate(axis=1, dtype=to_concat_chunks[0].dtype)
                concat = concat_op.new_chunk(
                    to_concat_chunks,
                    shape=(x_chunk.shape[0], concat_size),
                    order=TensorOrder.C_ORDER,
                    index=(i, 0),
                )

                chunk_op = op.copy().reset_key()
                chunk_op.stage = OperandStage.agg
                distance_params = {
                    "shape": (x_chunk.shape[0], k),
                    "order": TensorOrder.C_ORDER,
                    "dtype": np.dtype(np.float64),
                    "index": (i, 0),
                    "_type_": "distance",
                }
                if op.return_index:
                    index_params = {
                        "shape": (x_chunk.shape[0], k),
                        "order": TensorOrder.C_ORDER,
                        "dtype": np.dtype(np.int64),
                        "index": (i, 0),
                        "_type": "index",
                    }
                    distance_chunk, index_chunk = chunk_op.new_chunks(
                        [concat], kws=[distance_params, index_params]
                    )
                    out_distance_chunks.append(distance_chunk)
                    out_index_chunks.append(index_chunk)
                else:
                    out_distance_chunks.append(
                        chunk_op.new_chunk([concat], kws=[distance_params])
                    )

        new_op = op.copy()
        nsplits = (tuple(c.shape[0] for c in out_distance_chunks), (k,))
        params = [o.params for o in op.outputs]
        params[0]["chunks"] = out_distance_chunks
        params[0]["nsplits"] = nsplits
        if op.return_index:
            params[1]["chunks"] = out_index_chunks
            params[1]["nsplits"] = nsplits
        return new_op.new_tensors(op.inputs, kws=params)

    @classmethod
    def _topk_reduce_func(cls, dist, start, topk, xp, metric):
        """Reduce a chunk of distances to topk

        Parameters
        ----------
        dist : array of shape (n_samples_chunk, n_samples)
        start : int
            The index in X which the first row of dist corresponds to.
        topk : int

        Returns
        -------
        dist : array of shape (n_samples_chunk, n_neighbors)
        neigh : array of shape (n_samples_chunk, n_neighbors)
        """
        sample_range = xp.arange(dist.shape[0])[:, None]
        if topk - 1 >= dist.shape[1]:
            neigh_ind = xp.repeat(
                xp.arange(dist.shape[1]).reshape(1, -1), dist.shape[0], axis=0
            )
        else:
            neigh_ind = xp.argpartition(dist, topk - 1, axis=1)
            neigh_ind = neigh_ind[:, :topk]
        # argpartition doesn't guarantee sorted order, so we sort again
        neigh_ind = neigh_ind[sample_range, xp.argsort(dist[sample_range, neigh_ind])]
        return dist[sample_range, neigh_ind], neigh_ind

    @classmethod
    def _calcuate_topk_distances(cls, x, y, op, xp):
        metric = op.metric
        reduce_func = partial(cls._topk_reduce_func, topk=op.k, xp=xp, metric=op.metric)
        kwds = op.metric_kwargs or dict()
        need_sqrt = False
        if metric == "euclidean" and not kwds.get("squared", False):
            need_sqrt = True
            kwds["squared"] = True
        chunked_results = _pariwise_distance_chunked(
            x,
            y,
            reduce_func=reduce_func,
            metric=op.metric,
            working_memory=op.working_memory,
            xp=xp,
            **kwds
        )
        neigh_dist, neigh_ind = zip(*chunked_results)
        dist, ind = np.vstack(neigh_dist), np.vstack(neigh_ind)
        if metric == "euclidean" and need_sqrt:
            dist = xp.sqrt(dist)
        if getattr(op, "y_offset", None) is not None:
            ind += op.y_offset
        return dist, ind

    @classmethod
    def _execute_map(cls, ctx, op: "PairwiseDistancesTopk"):
        (x, y), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            xp = get_array_module(x, nosparse=True)
            ctx[op.outputs[0].key] = cls._calcuate_topk_distances(x, y, op, xp)

    @classmethod
    def _execute_agg(cls, ctx, op: "PairwiseDistancesTopk"):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )
        inputs = inputs[0]
        distances = inputs[0]

        with device(device_id):
            dist, ind = cls._topk_reduce_func(distances, 0, op.k, xp, op.metric)
            ctx[op.outputs[0].key] = dist
            if op.return_index:
                inds = inputs[1]
                ind_result = xp.empty_like(ind)
                for i in range(
                    len(ind_result)
                ):  # pylint: disable=consider-using-enumerate
                    ind_result[i] = inds[i][ind[i]]
                ctx[op.outputs[1].key] = ind_result

    @classmethod
    def _execute(cls, ctx, op: "PairwiseDistancesTopk"):
        (x, y), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            xp = get_array_module(x, nosparse=True)
            dist, ind = cls._calcuate_topk_distances(x, y, op, xp)
            dist, ind_on_ind = cls._topk_reduce_func(dist, 0, op.k, xp, op.metric)
            ctx[op.outputs[0].key] = dist
            if op.return_index:
                ind_result = xp.empty_like(ind_on_ind)
                for i in range(
                    len(ind_on_ind)
                ):  # pylint: disable=consider-using-enumerate
                    ind_result[i] = ind[i][ind_on_ind[i]]
                ctx[op.outputs[1].key] = ind_result

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            return cls._execute_map(ctx, op)
        elif op.stage == OperandStage.agg:
            return cls._execute_agg(ctx, op)
        else:
            return cls._execute(ctx, op)


def pairwise_distances_topk(
    X,
    Y=None,
    k=None,
    metric="euclidean",
    return_index=True,
    axis=1,
    working_memory=None,
    **kwds
):
    if k is None:  # pragma: no cover
        raise ValueError("`k` has to be specified")

    if Y is None:
        Y = X
    if axis == 0:
        X, Y = Y, X
    if working_memory is None:
        working_memory = options.learn.working_memory
    op = PairwiseDistancesTopk(
        x=X,
        y=Y,
        k=k,
        metric=metric,
        metric_kwargs=kwds,
        return_index=return_index,
        working_memory=working_memory,
    )
    return op(X, Y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ..datasource import tensor as astensor


def expand_dims(a, axis):
    """
    Expand the shape of a tensor.

    Insert a new axis that will appear at the `axis` position in the expanded
    array shape.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int
        Position in the expanded axes where the new axis is placed.

    Returns
    -------
    res : Tensor
        Output tensor. The number of dimensions is one greater than that of
        the input tensor.

    See Also
    --------
    squeeze : The inverse operation, removing singleton dimensions
    reshape : Insert, remove, and combine dimensions, and resize existing ones
    doc.indexing, atleast_1d, atleast_2d, atleast_3d

    Examples
    --------
    >>> import mars.tensor as mt

    >>> x = mt.array([1,2])
    >>> x.shape
    (2,)

    The following is equivalent to ``x[mt.newaxis,:]`` or ``x[mt.newaxis]``:

    >>> y = mt.expand_dims(x, axis=0)
    >>> y.execute()
    array([[1, 2]])
    >>> y.shape
    (1, 2)

    >>> y = mt.expand_dims(x, axis=1)  # Equivalent to x[:,mt.newaxis]
    >>> y.execute()
    array([[1],
           [2]])
    >>> y.shape
    (2, 1)

    Note that some examples may use ``None`` instead of ``np.newaxis``.  These
    are the same objects:

    >>> mt.newaxis is None
    True

    """
    a = astensor(a)

    if axis > a.ndim or axis < -a.ndim - 1:
        raise np.AxisError(
            f"Axis must be between -{a.ndim + 1} and {a.ndim}, got {axis}"
        )

    axis = axis if axis >= 0 else axis + a.ndim + 1
    indexes = (slice(None),) * axis + (np.newaxis,) + (slice(None),) * (a.ndim - axis)
    return a[indexes]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from typing import Dict, List

from ... import oscar as mo
from ...lib.aio import alru_cache
from ...resource import Resource
from ...storage import StorageLevel
from ...typing import BandType
from .core import NodeInfo, NodeStatus, WorkerSlotInfo, QuotaInfo, DiskInfo, StorageInfo
from .gather import gather_node_env, gather_node_resource, gather_node_details

logger = logging.getLogger(__name__)

DEFAULT_INFO_UPLOAD_INTERVAL = 1


class NodeInfoUploaderActor(mo.Actor):
    _band_slot_infos: Dict[str, List[WorkerSlotInfo]]
    _band_quota_infos: Dict[str, QuotaInfo]
    _disk_infos: List[DiskInfo]
    _band_storage_infos: Dict[str, Dict[StorageLevel, StorageInfo]]

    def __init__(self, role=None, interval=None, band_to_resource=None, use_gpu=True):
        self._info = NodeInfo(role=role)

        self._env_uploaded = False
        self._band_to_resource = band_to_resource

        self._interval = interval or DEFAULT_INFO_UPLOAD_INTERVAL
        self._upload_task = None
        self._upload_enabled = False
        self._uploaded_future = asyncio.Future()
        self._node_ready_event = asyncio.Event()

        self._use_gpu = use_gpu

        self._band_slot_infos = dict()
        self._band_quota_infos = dict()
        self._band_storage_infos = defaultdict(dict)
        self._disk_infos = []

    async def __post_create__(self):
        self._upload_task = asyncio.create_task(self._periodical_upload_node_info())
        await self._uploaded_future

    async def __pre_destroy__(self):
        self._upload_task.cancel()

    @alru_cache(cache_exceptions=False)
    async def _get_node_info_ref(self):
        from .locator import SupervisorLocatorActor
        from .supervisor.node_info import NodeInfoCollectorActor

        locator_ref = await mo.actor_ref(
            SupervisorLocatorActor.default_uid(), address=self.address
        )
        supervisor_addr = await locator_ref.get_supervisor(
            NodeInfoCollectorActor.default_uid()
        )
        if supervisor_addr is None:
            raise ValueError

        return await mo.actor_ref(
            NodeInfoCollectorActor.default_uid(), address=supervisor_addr
        )

    async def _periodical_upload_node_info(self):
        while True:
            try:
                await self.upload_node_info()
                if not self._uploaded_future.done():
                    self._uploaded_future.set_result(None)
            except asyncio.CancelledError:  # pragma: no cover
                break
            except Exception as ex:  # pragma: no cover  # noqa: E722  # nosec  # pylint: disable=bare-except
                logger.error(f"Failed to upload node info: {ex}")
                if not self._uploaded_future.done():
                    self._uploaded_future.set_exception(ex)
            try:
                await asyncio.sleep(self._interval)
            except asyncio.CancelledError:  # pragma: no cover
                break

    async def mark_node_ready(self):
        self._upload_enabled = True

        while True:
            try:
                # upload info in time to reduce latency
                await self.upload_node_info(status=NodeStatus.READY)
                break
            except (mo.ActorNotExist, ConnectionError):  # pragma: no cover
                await asyncio.sleep(1)

        self._node_ready_event.set()

    def is_node_ready(self):
        return self._node_ready_event.is_set()

    async def wait_node_ready(self):
        return self._node_ready_event.wait()

    async def upload_node_info(self, status: NodeStatus = None):
        try:
            if not self._info.env:
                self._info.env = await asyncio.to_thread(gather_node_env)
            self._info.detail.update(
                await asyncio.to_thread(
                    gather_node_details,
                    disk_infos=self._disk_infos,
                    band_storage_infos=self._band_storage_infos,
                    band_slot_infos=self._band_slot_infos,
                    band_quota_infos=self._band_quota_infos,
                )
            )

            band_resources = await asyncio.to_thread(
                gather_node_resource, self._band_to_resource, use_gpu=self._use_gpu
            )

            for band, res in band_resources.items():
                try:
                    res_dict = self._info.resource[band]
                except KeyError:
                    res_dict = self._info.resource[band] = dict()
                res_dict.update(res)

            if self._upload_enabled:
                try:
                    node_info_ref = await self._get_node_info_ref()
                    if not self._env_uploaded:
                        status = status or NodeStatus.READY
                    await node_info_ref.update_node_info(
                        address=self.address,
                        role=self._info.role,
                        env=self._info.env if not self._env_uploaded else None,
                        resource=self._info.resource,
                        detail=self._info.detail,
                        status=status,
                    )
                    self._env_uploaded = True
                except ValueError:
                    pass
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
            logger.exception(f"Failed to upload node info")
            raise

    def get_bands(self) -> Dict[BandType, int]:
        band_resource = dict()
        for resource_type, info in self._info.resource.items():
            if resource_type.startswith("numa"):
                # cpu
                band_resource[(self.address, resource_type)] = Resource(
                    num_cpus=info["cpu_total"], mem_bytes=info["memory_total"]
                )
            else:  # pragma: no cover
                assert resource_type.startswith("gpu")
                band_resource[(self.address, resource_type)] = Resource(
                    num_gpus=info["gpu_total"]
                )
        return band_resource

    def set_node_disk_info(self, node_disk_info: List[DiskInfo]):
        self._disk_infos = node_disk_info

    def set_band_storage_info(self, band_name: str, storage_info: StorageInfo):
        self._band_storage_infos[band_name][storage_info.storage_level] = storage_info

    def set_band_slot_infos(self, band_name, slot_infos: List[WorkerSlotInfo]):
        self._band_slot_infos[band_name] = slot_infos

    def set_band_quota_info(self, band_name, quota_info: QuotaInfo):
        self._band_quota_infos[band_name] = quota_info



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    from sklearn.metrics.pairwise import (
        manhattan_distances as sklearn_manhattan_distances,
    )
except ImportError:  # pragma: no cover
    sklearn_manhattan_distances = None

from .... import opcodes as OperandDef
from ....core import recursive_tile
from ....serialization.serializables import KeyField, BoolField
from ....tensor.core import TensorOrder
from ....tensor.arithmetic import abs as mt_abs
from ....tensor.spatial.distance import cdist
from ....tensor.array_utils import as_same_device, device
from ....utils import ensure_own_data
from .core import PairwiseDistances


class ManhattanDistances(PairwiseDistances):
    _op_type_ = OperandDef.PAIRWISE_MANHATTAN_DISTANCES

    _x = KeyField("x")
    _y = KeyField("y")
    _sum_over_features = BoolField("sum_over_features")

    def __init__(self, x=None, y=None, sum_over_features=None, use_sklearn=None, **kw):
        super().__init__(
            _x=x,
            _y=y,
            _sum_over_features=sum_over_features,
            _use_sklearn=use_sklearn,
            **kw,
        )

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def sum_over_features(self):
        return self._sum_over_features

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._x = self._inputs[0]
        self._y = self._inputs[1]

    def __call__(self, X, Y=None):
        X, Y = self.check_pairwise_arrays(X, Y)
        if self._y is None:
            self._y = Y

        if (X.issparse() or Y.issparse()) and not self._sum_over_features:
            raise TypeError(
                f"sum_over_features={self._sum_over_features} not supported"
                " for sparse matrices"
            )

        if not self._sum_over_features:
            shape = (X.shape[0] * Y.shape[0], X.shape[1])
        else:
            shape = (X.shape[0], Y.shape[0])

        return self.new_tensor([X, Y], shape=shape, order=TensorOrder.C_ORDER)

    @classmethod
    def tile(cls, op):
        x, y = op.x, op.y

        if len(x.chunks) == 1 and len(y.chunks) == 1:
            return cls._tile_one_chunk(op)

        if x.issparse() or y.issparse():
            assert op.sum_over_features
            return cls._tile_chunks(op, x, y)
        elif op.sum_over_features:
            # if x, y are not sparse and `sum_over_features` is True
            # just use cdist
            return [(yield from recursive_tile(cdist(x, y, "cityblock")))]
        else:
            d = x[:, np.newaxis, :] - y[np.newaxis, :, :]
            d = mt_abs(d)
            d = d.reshape((-1, x.shape[1]))
            return [(yield from recursive_tile(d))]

    @classmethod
    def execute(cls, ctx, op):
        (x, y), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )
        out = op.outputs[0]

        with device(device_id):
            if sklearn_manhattan_distances is not None:
                ctx[out.key] = sklearn_manhattan_distances(
                    ensure_own_data(x),
                    ensure_own_data(y),
                    sum_over_features=op.sum_over_features,
                )
            else:  # pragma: no cover
                # we cannot support sparse
                raise NotImplementedError(
                    "cannot support calculate manhattan distances on GPU"
                )


def manhattan_distances(X, Y=None, sum_over_features=True):
    """ Compute the L1 distances between the vectors in X and Y.

    With sum_over_features equal to False it returns the componentwise
    distances.

    Read more in the :ref:`User Guide <metrics>`.

    Parameters
    ----------
    X : array_like
        A tensor with shape (n_samples_X, n_features).

    Y : array_like, optional
        A tensor with shape (n_samples_Y, n_features).

    sum_over_features : bool, default=True
        If True the function returns the pairwise distance matrix
        else it returns the componentwise L1 pairwise-distances.
        Not supported for sparse matrix inputs.

    Returns
    -------
    D : Tensor
        If sum_over_features is False shape is
        (n_samples_X * n_samples_Y, n_features) and D contains the
        componentwise L1 pairwise-distances (ie. absolute difference),
        else shape is (n_samples_X, n_samples_Y) and D contains
        the pairwise L1 distances.

    Examples
    --------
    >>> from mars.learn.metrics.pairwise import manhattan_distances
    >>> manhattan_distances([[3]], [[3]]).execute() #doctest:+ELLIPSIS
    array([[0.]])
    >>> manhattan_distances([[3]], [[2]]).execute() #doctest:+ELLIPSIS
    array([[1.]])
    >>> manhattan_distances([[2]], [[3]]).execute() #doctest:+ELLIPSIS
    array([[1.]])
    >>> manhattan_distances([[1, 2], [3, 4]],\
         [[1, 2], [0, 3]]).execute() #doctest:+ELLIPSIS
    array([[0., 2.],
           [4., 4.]])
    >>> import mars.tensor as mt
    >>> X = mt.ones((1, 2))
    >>> y = mt.full((2, 2), 2.)
    >>> manhattan_distances(X, y, sum_over_features=False).execute() #doctest:+ELLIPSIS
    array([[1., 1.],
           [1., 1.]])
    """
    op = ManhattanDistances(
        x=X, y=Y, sum_over_features=sum_over_features, dtype=np.dtype(np.float64)
    )
    return op(X, Y=Y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from ..arithmetic.add import TensorTreeAdd
from .core import TensorCumReduction, TensorCumReductionMixin


class TensorCumsum(TensorCumReduction, TensorCumReductionMixin):
    _op_type_ = OperandDef.CUMSUM
    _func_name = "cumsum"

    def __init__(self, axis=None, **kw):
        super().__init__(_axis=axis, **kw)

    @staticmethod
    def _get_op_types():
        return TensorCumsum, TensorTreeAdd


def cumsum(a, axis=None, dtype=None, out=None):
    """
    Return the cumulative sum of the elements along a given axis.

    Parameters
    ----------
    a : array_like
        Input tensor.
    axis : int, optional
        Axis along which the cumulative sum is computed. The default
        (None) is to compute the cumsum over the flattened tensor.
    dtype : dtype, optional
        Type of the returned tensor and of the accumulator in which the
        elements are summed.  If `dtype` is not specified, it defaults
        to the dtype of `a`, unless `a` has an integer dtype with a
        precision less than that of the default platform integer.  In
        that case, the default platform integer is used.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must
        have the same shape and buffer length as the expected output
        but the type will be cast if necessary. See `doc.ufuncs`
        (Section "Output arguments") for more details.

    Returns
    -------
    cumsum_along_axis : Tensor.
        A new tensor holding the result is returned unless `out` is
        specified, in which case a reference to `out` is returned. The
        result has the same size as `a`, and the same shape as `a` if
        `axis` is not None or `a` is a 1-d tensor.


    See Also
    --------
    sum : Sum tensor elements.

    trapz : Integration of tensor values using the composite trapezoidal rule.

    diff :  Calculate the n-th discrete difference along given axis.

    Notes
    -----
    Arithmetic is modular when using integer types, and no error is
    raised on overflow.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.array([[1,2,3], [4,5,6]])
    >>> a.execute()
    array([[1, 2, 3],
           [4, 5, 6]])
    >>> mt.cumsum(a).execute()
    array([ 1,  3,  6, 10, 15, 21])
    >>> mt.cumsum(a, dtype=float).execute()     # specifies type of output value(s)
    array([  1.,   3.,   6.,  10.,  15.,  21.])

    >>> mt.cumsum(a,axis=0).execute()      # sum over rows for each of the 3 columns
    array([[1, 2, 3],
           [5, 7, 9]])
    >>> mt.cumsum(a,axis=1).execute()      # sum over columns for each of the 2 rows
    array([[ 1,  3,  6],
           [ 4,  9, 15]])

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.empty((1,), dtype=a.dtype).cumsum().dtype
    op = TensorCumsum(axis=axis, dtype=dtype)
    return op(a, out=out)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from typing import Any, List, Type, TypeVar

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....storage.base import StorageLevel, StorageFileObject
from ...cluster import StorageInfo
from ..core import (
    StorageManagerActor,
    DataManagerActor,
    DataInfo,
    WrappedStorageFileObject,
)
from ..handler import StorageHandlerActor
from .core import AbstractStorageAPI

_is_windows = sys.platform.lower().startswith("win")
APIType = TypeVar("APIType", bound="StorageAPI")


class StorageAPI(AbstractStorageAPI):
    _storage_handler_ref: mo.ActorRefType[StorageHandlerActor]
    _data_manager_ref: mo.ActorRefType[DataManagerActor]

    def __init__(self, address: str, session_id: str, band_name: str):
        self._address = address
        self._session_id = session_id
        self._band_name = band_name

    async def _init(self):
        self._storage_handler_ref = await mo.actor_ref(
            self._address, StorageHandlerActor.gen_uid(self._band_name)
        )
        self._data_manager_ref = await mo.actor_ref(
            self._address, DataManagerActor.default_uid()
        )

    @classmethod
    @alru_cache(cache_exceptions=False)
    async def create(
        cls: Type[APIType],
        session_id: str,
        address: str,
        band_name: str = "numa-0",
        **kwargs,
    ) -> APIType:
        """
        Create storage API.

        Parameters
        ----------
        session_id: str
            session id

        address: str
            worker address

        band_name: str
            name of band, default as 'numa-0'

        Returns
        -------
        storage_api
            Storage api.
        """
        if kwargs:  # pragma: no cover
            raise TypeError(f'Got unexpected arguments: {",".join(kwargs)}')
        api = StorageAPI(address, session_id, band_name)
        await api._init()
        return api

    @mo.extensible
    async def get(
        self, data_key: str, conditions: List = None, error: str = "raise"
    ) -> Any:
        return await self._storage_handler_ref.get(
            self._session_id, data_key, conditions, error
        )

    @get.batch
    async def batch_get(self, args_list, kwargs_list):
        gets = []
        for args, kwargs in zip(args_list, kwargs_list):
            gets.append(
                self._storage_handler_ref.get.delay(self._session_id, *args, **kwargs)
            )
        return await self._storage_handler_ref.get.batch(*gets)

    @mo.extensible
    async def put(
        self, data_key: str, obj: object, level: StorageLevel = None
    ) -> DataInfo:
        return await self._storage_handler_ref.put(
            self._session_id, data_key, obj, level
        )

    @put.batch
    async def batch_put(self, args_list, kwargs_list):
        puts = []
        for args, kwargs in zip(args_list, kwargs_list):
            puts.append(
                self._storage_handler_ref.put.delay(self._session_id, *args, **kwargs)
            )
        return await self._storage_handler_ref.put.batch(*puts)

    @mo.extensible
    async def get_infos(self, data_key: str) -> List[DataInfo]:
        """
        Get data information items for specific data key

        Parameters
        ----------
        data_key

        Returns
        -------
        out
            List of information for specified key
        """
        return await self._data_manager_ref.get_data_infos(
            self._session_id, data_key, self._band_name
        )

    @mo.extensible
    async def delete(self, data_key: str, error: str = "raise"):
        """
        Delete object.

        Parameters
        ----------
        data_key: str
            object key to delete
        error: str
            raise or ignore
        """
        await self._storage_handler_ref.delete(self._session_id, data_key, error=error)

    @delete.batch
    async def batch_delete(self, args_list, kwargs_list):
        deletes = []
        for args, kwargs in zip(args_list, kwargs_list):
            deletes.append(
                self._storage_handler_ref.delete.delay(
                    self._session_id, *args, **kwargs
                )
            )
        return await self._storage_handler_ref.delete.batch(*deletes)

    @mo.extensible
    async def fetch(
        self,
        data_key: str,
        level: StorageLevel = None,
        band_name: str = None,
        remote_address: str = None,
        error: str = "raise",
    ):
        """
        Fetch object from remote worker or load object from disk.

        Parameters
        ----------
        data_key: str
            data key to fetch to current worker with specific level
        level: StorageLevel
            the storage level to put into, MEMORY as default
        band_name: BandType
            put data on specific band
        remote_address:
            remote address that stores the data
        error: str
            raise or ignore
        """
        fetch_key = await self._storage_handler_ref.fetch_batch(
            self._session_id, [data_key], level, band_name, remote_address, error
        )
        if fetch_key:
            return fetch_key[0]

    @fetch.batch
    async def batch_fetch(self, args_list, kwargs_list):
        extracted_args = []
        data_keys = []
        for args, kwargs in zip(args_list, kwargs_list):
            data_key, level, band_name, dest_address, error = self.fetch.bind(
                *args, **kwargs
            )
            if extracted_args:
                assert extracted_args == (level, band_name, dest_address, error)
            extracted_args = (level, band_name, dest_address, error)
            data_keys.append(data_key)
        return await self._storage_handler_ref.fetch_batch(
            self._session_id, data_keys, *extracted_args
        )

    @mo.extensible
    async def unpin(self, data_key: str, error: str = "raise"):
        """
        Unpin the data, allow storage to release the data.

        Parameters
        ----------
        data_key: str
            data key to unpin
        error: str
            raise or ignore
        """
        await self._storage_handler_ref.unpin(self._session_id, data_key, error)

    @unpin.batch
    async def batch_unpin(self, args_list, kwargs_list):
        unpins = []
        for args, kwargs in zip(args_list, kwargs_list):
            data_key, error = self.unpin.bind(*args, **kwargs)
            unpins.append(
                self._storage_handler_ref.unpin.delay(self._session_id, data_key, error)
            )
        return await self._storage_handler_ref.unpin.batch(*unpins)

    async def open_reader(self, data_key: str) -> StorageFileObject:
        """
        Return a file-like object for reading.

        Parameters
        ----------
        data_key: str
            data key

        Returns
        -------
            return a file-like object.
        """
        return await self._storage_handler_ref.open_reader(self._session_id, data_key)

    async def open_writer(
        self, data_key: str, size: int, level: StorageLevel
    ) -> WrappedStorageFileObject:
        """
        Return a file-like object for writing data.

        Parameters
        ----------
        data_key: str
            data key
        size: int
            the total size of data
        level: StorageLevel
            the storage level to write

        Returns
        -------
            return a file-like object.
        """
        return await self._storage_handler_ref.open_writer(
            self._session_id, data_key, size, level
        )

    async def list(self, level: StorageLevel) -> List:
        """
        List all stored data_keys in storage.

        Parameters
        ----------
        level: StorageLevel
            the storage level to list all objects

        Returns
        -------
            list of data keys
        """
        return await self._storage_handler_ref.list(level=level)

    async def get_storage_level_info(self, level: StorageLevel) -> StorageInfo:
        """
        Get storage level's info.

        Parameters
        ----------
        level : StorageLevel
            Storage level.

        Returns
        -------
        storage_level_info : StorageInfo
        """
        return await self._storage_handler_ref.get_storage_level_info(level)

    async def get_storage_info(self, level: StorageLevel) -> dict:
        """
        Get the customized storage backend info of requested storage backend.

        Parameters
        ----------
        level: StorageLevel
            the storage level to fetch the backend info.

        Returns
        -------
        info : dict
            Customized storage backend info dict.
        """
        return await self._storage_handler_ref.get_storage_backend_info(level)


class MockStorageAPI(StorageAPI):
    @classmethod
    async def create(
        cls: Type[APIType], session_id: str, address: str, **kwargs
    ) -> APIType:
        from ..core import StorageManagerActor

        storage_configs = kwargs.get("storage_configs")
        if not storage_configs:
            if sys.platform == "darwin":
                plasma_dir = "/tmp"
            else:
                plasma_dir = "/dev/shm"
            plasma_setup_params = dict(
                store_memory=10 * 1024 * 1024,
                plasma_directory=plasma_dir,
                check_dir_size=False,
            )
            if _is_windows:
                storage_configs = {"shared_memory": {}}
            else:
                storage_configs = {
                    "plasma": plasma_setup_params,
                }

        storage_handler_cls = kwargs.pop("storage_handler_cls", StorageHandlerActor)
        await mo.create_actor(
            StorageManagerActor,
            storage_configs,
            storage_handler_cls=storage_handler_cls,
            uid=StorageManagerActor.default_uid(),
            address=address,
        )
        return await super().create(address=address, session_id=session_id)

    @classmethod
    async def cleanup(cls: Type[APIType], address: str):
        await mo.destroy_actor(
            await mo.actor_ref(address, StorageManagerActor.default_uid())
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from .custom_log import CustomLogActor


class SessionWorkerService(AbstractService):
    """
    Session service on worker.

    Service Configuration
    ---------------------
    {
        "session" : {
        }
    }
    """

    async def start(self):
        session_config = self._config.get("session", dict())
        custom_log_dir = session_config.get("custom_log_dir")
        await mo.create_actor(
            CustomLogActor,
            custom_log_dir,
            address=self._address,
            uid=CustomLogActor.default_uid(),
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(uid=CustomLogActor.default_uid(), address=self._address)
        )
